# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.layers import layer_utils
from neurst.layers.search import register_search_layer
from neurst.layers.search.sequence_search import SequenceSearch
from neurst.utils import compat
from neurst.utils.flags_core import Flag


class _StateKeys(object):
    """Keys to dictionary storing the state of the sampling search loop."""

    # The decoding step.
    TIME_STEP = "TIME_STEP"

    # The decoder input ids with shape [batch_size * sample_num, ]
    INPUT_IDS = "INPUT_IDS"

    # Dictionary of cached values for each beam. The cache stores encoder
    # output (memory), encoder input paddings and the decoder attention
    # output from the previous iteration (usually for transformer)
    CACHE = "CACHE"

    # Flags indicating which sequences in the finished sequences are finished,
    # with shape [batch_size * sample_num, ]
    FINISHED_FLAGS = "FINISHED_FLAGS"

    # The true decoding length of each beam
    DECODING_LENGTH = "LENGTH"

    # The predicted sequences, with shape [time_step, batch_size * sample_num]
    PREDICTED_IDS = "PREDICTED_IDS"


def _calculate_logits(state, symbols_to_logits_fn, unk_id):
    """ Calculates one-step logits. UNK will be masked.

    Args:
        state: A dictionary containing current state of beam search.
        symbols_to_logits_fn:
        unk_id: An int scalar, indicating the unknown token id.

    Returns:
        A float tensor with the same shape as `logits`.
    """
    logits = symbols_to_logits_fn(
        state[_StateKeys.INPUT_IDS], state[_StateKeys.CACHE],
        state[_StateKeys.TIME_STEP])

    vocab_size = logits.get_shape().as_list()[-1]
    unk_bias = layer_utils.one_entry_bias(
        on_entry=unk_id, num_entries=vocab_size,
        on_value=compat.FLOAT_MIN, off_value=0.0, dtype=logits.dtype)
    unk_coefficient = layer_utils.one_entry_bias(
        on_entry=unk_id, num_entries=vocab_size,
        on_value=0.0, off_value=1.0, dtype=logits.dtype)

    logits = unk_coefficient * logits + unk_bias

    return logits


def top_k_logits(logits, k):
    if k == 0:
        # no truncation
        return logits

    values, _ = tf.nn.top_k(logits, k=k)
    min_values = values[:, -1, tf.newaxis]
    return tf.where(logits < min_values,
                    tf.ones_like(logits, dtype=logits.dtype) * compat.FLOAT_MIN, logits)


def top_p_logits(logits, p):
    sorted_logits = tf.sort(logits, direction="DESCENDING")
    cumulative_probs = tf.cumsum(tf.nn.softmax(sorted_logits), axis=-1)
    t_sorted_indices_to_remove = cumulative_probs < p
    min_value = tf.reduce_min(logits)
    applyed_sorted_logits = tf.where(
        t_sorted_indices_to_remove,
        tf.ones_like(sorted_logits, dtype=sorted_logits.dtype) * min_value,
        sorted_logits,
    )
    threshold = tf.expand_dims(tf.reduce_max(applyed_sorted_logits, axis=-1), 1)
    return tf.where(
        logits < threshold, tf.ones_like(logits, dtype=logits.dtype) * compat.FLOAT_MIN, logits
    )


def sequence_sampling_search(symbols_to_logits_fn,
                             generation_initializer,
                             sample_next_word_fn,
                             sample_num,
                             extra_decode_length=50,
                             maximum_decode_length=256,
                             minimum_decode_length=0,
                             padded_decode=False,
                             dtype=None, ):
    """ Search for target subtokens with the largest probability.

    Args:
        symbols_to_logits_fn: A callable function that takes ids, cache and current
            timestep as arguments. The passed in arguments will have following shape:
                ids -> A tensor with shape [batch_size * beam_size,].
                cache -> A nested dictionary of tensors [batch_size * beawm_size, ...].
                index -> A scalar.
            The function must return a logits tensor with shape
                [batch_size * beam_size, vocab_size].
        generation_initializer: A dict from model, containing:
            -> decoder_input_ids: An int32 tensor with shape [batch_size,].
                    The initial input for the decoder.
            -> decoding_cache: A nested dictionary, containing attention memory and other
                    starting variables information.
            -> encoder_inputs_maxlen: An int32 scalar tensor, indicating the maximum length of encoding input
                    sequence.
            -> eos_id: An int scalar, indicating the end-of-sentence symbol id, used to determine when a
                    sequence has finished.
            -> unk_id: An int scalar, indicating the unknown token id.
        sample_next_word_fn:
        sample_num:
        extra_decode_length: An int scalar. The real search steps will be
            `encoder_inputs_maxlen` + `extra_decode_length`.
        maximum_decode_length: An int scalar, indicating the maximum decode length. Prediction
            outputs will be padded to this length.
        minimum_decode_length: An int scalar, indicating the minimum decode length.
        padded_decode: whether the autoregressive decoding runs with input data padded to the decode_max_length.
        dtype: A string or a tensorflow data type used for score computation. If None, auto set
            to the GLOBAL_FLOATX.

    Returns:
        A tuple of (hypothesis, scores):
            hypothesis: int32 tensor with shape [batch_size * top_k, decode_length]
            scores: float tensor with shape [batch_size * top_k]
    """
    _ = dtype
    decoder_input_ids = generation_initializer["decoder_input"]
    decoding_cache = generation_initializer["decoder_internal_cache"]
    encoder_inputs_maxlen = generation_initializer["encoder_inputs_maxlen"]
    eos_id = generation_initializer["eos_id"]
    unk_id = generation_initializer["unk_id"]
    batch_size = tf.shape(decoder_input_ids)[0]
    initial_finished = tf.tile([False], [batch_size * sample_num])
    initial_decoder_input_ids = layer_utils.stack_beam_size(decoder_input_ids, sample_num)
    initial_time = tf.constant(0, dtype=tf.int32)
    initial_cache = tf.nest.map_structure(
        lambda x: layer_utils.stack_beam_size(x, sample_num), decoding_cache
    )
    # [time, batch_size * sample_num]
    if padded_decode:
        initial_predicted_ids = tf.zeros([batch_size * sample_num, maximum_decode_length],
                                         dtype=tf.int32)
    else:
        initial_predicted_ids = tf.zeros([batch_size * sample_num, 0],
                                         dtype=tf.int32)
    # Create state dictionary
    search_state = {
        _StateKeys.TIME_STEP: initial_time,
        _StateKeys.INPUT_IDS: tf.cast(initial_decoder_input_ids, tf.int32),
        _StateKeys.CACHE: initial_cache,
        _StateKeys.FINISHED_FLAGS: initial_finished,
        _StateKeys.PREDICTED_IDS: initial_predicted_ids,
    }
    # Create state invariants for each value in the state dictionary. Each
    # dimension must be a constant or None. A None dimension means either:
    #   1) the dimension's value is a tensor that remains the same but may
    #      depend on the input sequence to the model (e.g. batch size).
    #   2) the dimension may have different values on different iterations.
    if padded_decode:
        search_state_shape_invariants = {
            _StateKeys.TIME_STEP: tf.TensorShape([]),
            _StateKeys.INPUT_IDS: tf.TensorShape([None]),
            _StateKeys.CACHE: tf.nest.map_structure(
                layer_utils.static_tensorshape, initial_cache
            ),
            _StateKeys.FINISHED_FLAGS: tf.TensorShape([None]),
            _StateKeys.PREDICTED_IDS: tf.TensorShape([None, maximum_decode_length]),
        }
    else:
        search_state_shape_invariants = {
            _StateKeys.TIME_STEP: tf.TensorShape([]),
            _StateKeys.INPUT_IDS: tf.TensorShape([None]),
            _StateKeys.CACHE: tf.nest.map_structure(
                layer_utils.dynamic_tensorshape_except_last_dim, initial_cache
            ),
            _StateKeys.FINISHED_FLAGS: tf.TensorShape([None]),
            _StateKeys.PREDICTED_IDS: tf.TensorShape([None, None]),
        }

    maximum_search_steps = tf.minimum(
        encoder_inputs_maxlen + extra_decode_length, maximum_decode_length)
    maximum_search_steps = tf.maximum(maximum_search_steps, minimum_decode_length)

    def search_step(state):
        """ Beam search step. """
        # [batch_size * beam_size, vocab_size]
        logits = _calculate_logits(
            state=state, unk_id=unk_id,
            symbols_to_logits_fn=symbols_to_logits_fn)
        # filter out eos
        vocab_size = logits.get_shape().as_list()[-1]
        eos_bias = layer_utils.one_entry_bias(
            on_entry=eos_id, num_entries=vocab_size,
            on_value=compat.FLOAT_MIN, off_value=0.0, dtype=logits.dtype)
        eos_coefficient = layer_utils.one_entry_bias(
            on_entry=eos_id, num_entries=vocab_size,
            on_value=0.0, off_value=1.0, dtype=logits.dtype)
        logits = tf.cond(tf.less(state[_StateKeys.TIME_STEP], minimum_decode_length - 1),
                         lambda: eos_coefficient * logits + eos_bias,
                         lambda: logits)

        # compute log probs and generate next token ids according to beam scores
        logits = sample_next_word_fn(logits=logits)
        # [batch*beam, 1]
        sample_ids = tf.random.categorical(logits, num_samples=1, dtype=tf.int32)
        next_predicted_ids = state[_StateKeys.PREDICTED_IDS]
        if padded_decode:
            next_predicted_ids = tf.transpose(tf.tensor_scatter_nd_update(
                tf.transpose(next_predicted_ids), [[state[_StateKeys.TIME_STEP]]],
                tf.transpose(sample_ids)))
        else:
            next_predicted_ids = tf.concat([next_predicted_ids, sample_ids], axis=1)

        next_cache = state[_StateKeys.CACHE]
        pre_finish = state[_StateKeys.FINISHED_FLAGS]
        cur_finished = tf.equal(eos_id, tf.squeeze(sample_ids, 1))
        next_finished = tf.where(pre_finish, True, cur_finished)
        state.update({
            _StateKeys.TIME_STEP: state[_StateKeys.TIME_STEP] + 1,
            _StateKeys.INPUT_IDS: tf.squeeze(sample_ids, 1),
            _StateKeys.CACHE: next_cache,
            _StateKeys.FINISHED_FLAGS: next_finished,
            _StateKeys.PREDICTED_IDS: next_predicted_ids})
        return [state]

    # beam search step by step
    res = tf.while_loop(
        lambda state: tf.logical_and(
            tf.logical_not(tf.reduce_all(state[_StateKeys.FINISHED_FLAGS])),
            tf.less(state[_StateKeys.TIME_STEP], maximum_search_steps)),
        search_step,
        loop_vars=[search_state],
        shape_invariants=[search_state_shape_invariants],
        parallel_iterations=1,
    )
    res = res[0]
    hypothesis = res[_StateKeys.PREDICTED_IDS]
    # padding to fixed length output
    if not padded_decode:
        hypothesis = tf.pad(
            hypothesis,
            paddings=tf.convert_to_tensor(
                [[0, 0], [0, maximum_decode_length - res[_StateKeys.TIME_STEP]]],
                dtype=tf.int32),
            mode="CONSTANT",
            constant_values=eos_id)

    return hypothesis


@register_search_layer("TopSampling")
class Sampling(SequenceSearch):
    def __init__(self, args):
        super(Sampling, self).__init__()
        self._top_p = args["top_p"]
        self._top_k = args["top_k"]
        self._sample_num = args["sample_num"]
        self._maximum_decode_length = args["maximum_decode_length"]
        self._extra_decode_length = args["extra_decode_length"]
        self._minimum_decode_length = args["minimum_decode_length"]
        if not self._minimum_decode_length:
            self._minimum_decode_length = 0
        if self._top_p < 1 and self._top_k > 0:
            raise NotImplementedError("Not implemented search logic when top_k > 0 and top_p < 1.")
        self._padded_decode = args["padded_decode"]

    @staticmethod
    def class_or_method_args():
        return [
            Flag("sample_num", dtype=Flag.TYPE.INTEGER,
                 default=1, help="The number of copies for each input item.", ),
            Flag("top_k", dtype=Flag.TYPE.INTEGER,
                 default=0, help="The number of token in each step for top_k sampling."),
            Flag("top_p", dtype=Flag.TYPE.FLOAT,
                 default=1.0, help="The threshold for cumulated probability."),
            Flag("maximum_decode_length", dtype=Flag.TYPE.INTEGER,
                 default=None, help="The maximum decoding length of sampling."),
            Flag("minimum_decode_length", dtype=Flag.TYPE.INTEGER,
                 default=0, help="The minimum decoding length of sampling."),
            Flag("extra_decode_length", dtype=Flag.TYPE.INTEGER, default=50,
                 help="The extra decoding length versus source side for beam search inference. "
                      "The maximum decoding length of beam search inference will be "
                      "source_sequence_length + extra_decode_length if maximum_decode_length "
                      "if not provided."),
            Flag("padded_decode", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether the autoregressive decoding runs with input data padded to "
                      "the decode_max_length. For TPU/XLA-GPU runs, this flag has to be "
                      "set due the static shape requirement. In addition, this method "
                      "will introduce unnecessary overheads which grow quadratically with "
                      "the max sequence length."),
        ]

    def call(self, parsed_inp, **kwargs):
        """ Do beam search.

        Args:
            parsed_inp: A dict of parsed model inputs.

        Returns:
            The search results (must be a tuple).
        """
        maximum_decode_length = self._maximum_decode_length
        max_data_len = kwargs.get("max_data_len", None)
        if maximum_decode_length is None:
            assert max_data_len, (
                "`maximum_decode_length` must be provided " "when `max_data_len` is not provided.")
            maximum_decode_length = self._extra_decode_length + max_data_len
        if self._minimum_decode_length >= maximum_decode_length:
            raise ValueError("`minimum_decode_length` must be less than maximum decode length.")

        with tf.name_scope("top_sampling"):
            decode_padded_length = self._maximum_decode_length if self._padded_decode else None
            symbols_to_logits_fn, generation_initializer = self._model.get_symbols_to_logits_fn(
                parsed_inp, is_training=False, is_inference=True,
                decode_padded_length=decode_padded_length)

        def _sample_fn(logits):
            if self._top_p < 1:
                return top_p_logits(logits, self._top_p)
            return top_k_logits(logits, self._top_k)

        hypothesis = sequence_sampling_search(
            symbols_to_logits_fn,
            generation_initializer,
            sample_next_word_fn=_sample_fn,
            sample_num=self._sample_num,
            extra_decode_length=self._extra_decode_length,
            maximum_decode_length=self._maximum_decode_length,
            minimum_decode_length=self._minimum_decode_length,
            padded_decode=self._padded_decode)

        return hypothesis,



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf

from neurst.utils.configurable import ModelConfigs


@six.add_metaclass(ABCMeta)
class Converter(object):
    """ Abstract class for converting models and tasks. """
    REGISTRY_NAME = "converter"

    @classmethod
    def new(cls, *args, **kwargs):
        _ = args
        _ = kwargs
        return cls

    @staticmethod
    @abstractmethod
    def convert_model_config(path):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def convert_task_config(path):
        raise NotImplementedError

    @staticmethod
    def download(key):
        _ = key
        return None

    @staticmethod
    @abstractmethod
    def convert_checkpoint(path, save_path):
        raise NotImplementedError

    @classmethod
    def convert(cls, from_path, to_path):
        if not tf.io.gfile.exists(from_path):
            path = cls.download(from_path)
            if path is None:
                raise ValueError(f"Fail to find model to download: {from_path}")
            from_path = path
        try:
            cfgs = cls.convert_model_config(from_path)
        except NotImplementedError:
            cfgs = {}
        try:
            cfgs.update(cls.convert_task_config(from_path))
        except NotImplementedError:
            pass
        ModelConfigs.dump(cfgs, to_path)
        cls.convert_checkpoint(from_path, to_path)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import tensorflow as tf
from absl import logging

from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer


@register_tokenizer("spm")
class SentencePiece(Tokenizer):

    def __init__(self, language, glossaries=None, **kwargs):
        _ = kwargs
        super(SentencePiece, self).__init__(
            language=language, glossaries=glossaries)
        try:
            import sentencepiece
            self._sp = sentencepiece.SentencePieceProcessor()
        except ImportError:
            raise ImportError('Please install SentencePiece with: pip install sentencepiece')
        self._built = False
        self._codes = None

    def _lazy_init(self):
        codes = self._codes
        if codes.startswith("hdfs://"):
            local_path = os.path.join(os.path.dirname(__file__), "spm.model")
            logging.info("Copying spm model: {} to local: {}".format(codes, local_path))
            tf.io.gfile.copy(codes, local_path, overwrite=True)
            codes = local_path
        status = self._sp.Load(codes)
        assert status, "Fail to load spm model: {}".format(codes)
        self._built = True

    def init_subtokenizer(self, codes):
        """ Lazily initializes sentence piece. """
        self._codes = codes

    def tokenize(self, text, return_str=False):
        if not self._built:
            self._lazy_init()
            if not self._built:
                raise ValueError("call `init_subtokenizer` at first to initialize the SentencePiece.")
        return self._output_wrapper(
            self._sp.EncodeAsPieces(self._convert_to_str(text)), return_str=return_str)

    def detokenize(self, text, return_str=True):
        if not self._built:
            self._lazy_init()
            if not self._built:
                raise ValueError("call `init_subtokenizer` at first to initialize the SentencePiece.")
        return self._output_wrapper(
            self._sp.DecodePieces(self._convert_to_list(text)), return_str=return_str)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
import zipfile

import numpy
import tensorflow as tf
import yaml
from absl import logging

from neurst.models.bert import Bert
from neurst.utils.converters import Converter, register_converter
from neurst.utils.misc import download_with_tqdm

_URL_PREFIX = "https://storage.googleapis.com/bert_models"
_BERT_PRETRAIN_MODELS = {
    "bert-base-uncased": _URL_PREFIX + "/2018_10_18/uncased_L-12_H-768_A-12.zip",
    # "bert-base-uncased": _URL_PREFIX + "/2020_02_20/uncased_L-12_H-768_A-12.zip"
    "bert-large-uncased": _URL_PREFIX + "/2018_10_18/uncased_L-24_H-1024_A-16.zip",
    "bert-base-chinese": _URL_PREFIX + "/2018_11_03/chinese_L-12_H-768_A-12.zip",

}

_DIRECT_MAPPINGS = {
    "bert/embeddings/word_embeddings": "bert/bert_embedding/word_embedding",
    "bert/embeddings/token_type_embeddings": "bert/bert_embedding/token_type_embedding",
    "bert/embeddings/position_embeddings": "bert/bert_embedding/position_embedding",
    "bert/pooler/dense/bias": "bert/pooler/bias",
    "bert/pooler/dense/kernel": "bert/pooler/kernel",
    "bert/embeddings/LayerNorm/beta": "bert/bert_embedding/ln/beta",
    "bert/embeddings/LayerNorm/gamma": "bert/bert_embedding/ln/gamma",
}

_POSTFIX_MAPPINGS = {
    "attention/output/dense/kernel": "self_attention_prepost_wrapper/self_attention/output_transform/kernel",
    "attention/output/dense/bias": "self_attention_prepost_wrapper/self_attention/output_transform/bias",
    "attention/output/LayerNorm/beta": "self_attention_prepost_wrapper/ln/beta",
    "attention/output/LayerNorm/gamma": "self_attention_prepost_wrapper/ln/gamma",
    "intermediate/dense/kernel": "ffn_prepost_wrapper/ffn/dense1/kernel",
    "intermediate/dense/bias": "ffn_prepost_wrapper/ffn/dense1/bias",
    "output/dense/kernel": "ffn_prepost_wrapper/ffn/dense2/kernel",
    "output/dense/bias": "ffn_prepost_wrapper/ffn/dense2/bias",
    "output/LayerNorm/beta": "ffn_prepost_wrapper/ln/beta",
    "output/LayerNorm/gamma": "ffn_prepost_wrapper/ln/gamma",
}


@register_converter
class GoogleBert(Converter):
    """ Converts from the google bert.
    https://github.com/google-research/bert
    """

    @staticmethod
    def convert_model_config(path):
        with tf.io.gfile.GFile(os.path.join(path, "bert_config.json")) as fp:
            cfg = yaml.load(fp, Loader=yaml.FullLoader)
        return {
            "model.class": Bert.__name__,
            "model.params": {
                "max_position_embeddings": cfg["max_position_embeddings"],
                "num_layers": cfg["num_hidden_layers"],
                "hidden_size": cfg["hidden_size"],
                "num_attention_heads": cfg["num_attention_heads"],
                "filter_size": cfg["intermediate_size"],
                "ffn_activation": cfg["hidden_act"],
                "attention_dropout_rate": cfg["attention_probs_dropout_prob"],
                "attention_type": "dot_product",
                "ffn_dropout_rate": cfg["hidden_dropout_prob"],
                "layer_postprocess_dropout_rate": cfg["hidden_dropout_prob"]
            }
        }

    @staticmethod
    def convert_task_config(path):
        raise NotImplementedError

    @staticmethod
    def download(key):
        if key in _BERT_PRETRAIN_MODELS:
            url = _BERT_PRETRAIN_MODELS[key]
        elif key.startswith("http://") or key.startswith("https://"):
            url = key
        else:
            return None
        bert_name = os.path.basename(url).split(".")[0]
        this_dir = os.path.dirname(__file__)
        extract_path = os.path.join(this_dir, bert_name)
        if not os.path.exists(extract_path):
            logging.info(f"Downloading google bert: {key}")
            tarball = os.path.join(this_dir, os.path.basename(url))
            download_with_tqdm(url, tarball)
            tf.io.gfile.makedirs(extract_path)
            with zipfile.ZipFile(tarball) as zip_ref:
                zip_ref.extractall(extract_path)
            os.remove(tarball)
        if os.path.isdir(os.path.join(extract_path, bert_name)):
            return os.path.join(extract_path, bert_name)
        return extract_path

    @staticmethod
    def convert_checkpoint(path, save_path):
        ckpt = os.path.join(path, "bert_model.ckpt")
        bert_var_names = [x[0] for x in tf.train.list_variables(ckpt)]
        new_vars = []
        processed_var_names = []
        for var_name in bert_var_names:
            var_value = tf.train.load_variable(ckpt, var_name)
            new_var_name = None
            if var_name in _DIRECT_MAPPINGS:
                new_var_name = _DIRECT_MAPPINGS[var_name]
            elif var_name.startswith("bert/encoder/layer_"):
                lid = re.search(r"layer_\d+", var_name).group().split("_")[-1]
                postfix = var_name.split(f"layer_{lid}/")[1]
                if postfix in _POSTFIX_MAPPINGS:
                    new_var_name = f"bert/encoder/layer_{lid}/{_POSTFIX_MAPPINGS[postfix]}"
                elif (postfix.startswith("attention/self/key")
                      or postfix.startswith("attention/self/query")
                      or postfix.startswith("attention/self/value")):
                    tensor_name = postfix.split("/")[-1]
                    new_var_name = (f"bert/encoder/layer_{lid}/self_attention_prepost_wrapper"
                                    f"/self_attention/qkv_transform/{tensor_name}")
                    if new_var_name in processed_var_names:
                        continue
                    q_value = tf.train.load_variable(
                        ckpt, f"bert/encoder/layer_{lid}/attention/self/query/{tensor_name}")
                    k_value = tf.train.load_variable(
                        ckpt, f"bert/encoder/layer_{lid}/attention/self/key/{tensor_name}")
                    v_value = tf.train.load_variable(
                        ckpt, f"bert/encoder/layer_{lid}/attention/self/value/{tensor_name}")
                    var_value = numpy.concatenate([q_value, k_value, v_value], axis=-1)

            if new_var_name:
                processed_var_names.append(new_var_name)
                logging.info(f"convert {var_name}")
                logging.info(f"\t ==> {new_var_name}")
                new_vars.append(tf.Variable(
                    initial_value=var_value,
                    trainable=False,
                    name=new_var_name,
                    dtype=str(var_value.dtype)))
            else:
                logging.info(f"No matching variable for {var_name}")
        ckpt_saver = tf.train.Checkpoint(**{x.name.split(":")[0]: x for x in new_vars})
        save_ckpt = os.path.join(save_path, "ckpt")
        logging.info(f"Saving checkpoint to {save_ckpt}")
        ckpt_saver.save(save_ckpt)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import traceback
from abc import ABCMeta

import numpy
import six
import tensorflow as tf
from absl import logging
from tensorflow.python.keras import backend as K
from tensorflow.python.ops import summary_ops_v2

from neurst.layers.metric_layers import METRIC_REDUCTION, get_metric_reduction
from neurst.utils import compat
from neurst.utils.configurable import ModelConfigs


@six.add_metaclass(ABCMeta)
class CentralizedCallback(tf.keras.callbacks.Callback):
    """ Custom base Callback for handling global step. """

    def __init__(self):
        super(CentralizedCallback, self).__init__()
        self.__global_step = compat.get_registered_initial_step()

    def on_train_begin(self, logs=None):
        super(CentralizedCallback, self).on_train_begin(logs)
        self.__global_step = compat.get_registered_initial_step()

    def on_train_batch_begin(self, batch, logs=None):
        self.__global_step += 1
        self.custom_on_train_batch_begin(self.__global_step, logs)

    def on_train_batch_end(self, batch, logs=None):
        self.custom_on_train_batch_end(self.__global_step, logs)

    def custom_on_train_batch_begin(self, step, logs=None):
        pass

    def custom_on_train_batch_end(self, step, logs=None):
        pass


class CustomCheckpointCallback(CentralizedCallback):
    """ Defines custom checkpoint callback for automatically saving checkpoints.
        It DOES NOT support multi-worker mirrored strategy.
    """

    def __init__(self,
                 model_configs,
                 save_checkpoint_steps=1000,
                 checkpoint_manager=None):
        """ Initializes custom checkpoint callback.

        Args:
            model_configs: A dict of configurations for restoring.
            save_checkpoint_steps: An int scalar, saving checkpoint this every steps.
            checkpoint_manager: A CheckpointManager instance.
        """
        super(CustomCheckpointCallback, self).__init__()
        self._checkpoint_manager = checkpoint_manager
        if self._checkpoint_manager is None:
            self._checkpoint_manager = compat.get_saver_or_default()
        self._model_configs = model_configs
        self._save_checkpoint_steps = save_checkpoint_steps

    def on_train_begin(self, logs=None):
        super(CustomCheckpointCallback, self).on_train_begin(logs)
        ModelConfigs.dump(self._model_configs, output_dir=self._checkpoint_manager.directory)

    def custom_on_train_batch_end(self, step, logs=None):
        """ Save checkpoints. """
        _ = logs
        if step % self._save_checkpoint_steps == 0:
            start_time = time.time()
            try:
                path = self._checkpoint_manager.save(step)
            except tf.errors.OpError:
                logging.info(traceback.format_exc())
                logging.info("Fail to save checkpoint.")
            else:
                logging.info("Saved checkpoint into %s\tElapsed %.2fs", path, time.time() - start_time)


class LearningRateScheduler(CentralizedCallback):
    """ Custom learning rate scheduler which adapts by steps. """

    def __init__(self, lr_schedule):
        """ Initializes. """
        super(LearningRateScheduler, self).__init__()
        logging.info("Wrapper lr schedule with the callback LearningRateScheduler.")
        self._lr_schedule = lr_schedule

    def custom_on_train_batch_end(self, step, logs=None):
        if not hasattr(self.model.optimizer, 'lr'):
            raise ValueError('Optimizer must have a "lr" attribute.')
        lr = self._lr_schedule(step)
        if not isinstance(lr, (tf.Tensor, float, numpy.float32, numpy.float64)):
            raise ValueError('The output of the "schedule" function '
                             'should be float.')
        if isinstance(lr, tf.Tensor) and not lr.dtype.is_floating:
            raise ValueError('The dtype of Tensor should be float')
        K.set_value(self.model.optimizer.lr, K.get_value(lr))


class MetricReductionCallback(CentralizedCallback):

    def __init__(self, strategy, summary_steps, log_dir,
                 device='', lr_schedule=None):
        super(MetricReductionCallback, self).__init__()
        self._summary_steps = summary_steps
        self._log_dir = log_dir
        self._file_writer = tf.summary.create_file_writer(self._log_dir)
        self._file_writer.set_as_default()
        self._lr_schedule = lr_schedule
        self._should_summary = (compat.get_distributed_worker_setting()[0] == 0)
        # logging
        self._last_triggered_time = None
        self._accumulated_time_secs = 0.
        self._last_triggered_metric = None
        # for allreduce
        self._strategy = strategy
        self._device = device
        self._m_vars = {}
        self._allreduce_ops = {}
        self._allreduce_ranks = 1.

    def on_train_begin(self, logs=None):
        """ At the begining of training, write the graph to the tensorboard. """
        super(MetricReductionCallback, self).on_train_begin(logs)
        if self._should_summary:
            summary_ops_v2.graph(K.get_graph(), step=0)

    def custom_on_train_batch_begin(self, step, logs=None):
        self._last_triggered_time = time.time()

    def _byteps_average_metrics_in_place(self, logs):
        logs = logs or {}
        reduced_logs = {}
        import byteps.tensorflow as bps

        if self._allreduce_ranks <= 1.:
            self._allreduce_ranks = float(bps.size())
        # Reduce every metric among workers. Sort metrics by name
        # to ensure consistent order.
        for metric, value in sorted(logs.items()):
            from tensorflow.python.eager import context
            if context.executing_eagerly():
                with tf.device(self._device):
                    reduced_logs[metric] = bps.push_pull(K.constant(value, name=metric)).numpy()
            else:
                if metric not in self.variables:
                    with tf.name_scope('MetricAverageCallback') as scope:
                        var = tf.Variable(value, name=metric)
                        K.get_session().run(var.initializer)
                        self._m_vars[metric] = var
                        self._allreduce_ops[metric] = bps.push_pull(var, scope, device_dense=self._device)
                else:
                    K.set_value(self._m_vars[metric], value)
                reduced_logs[metric] = K.get_session().run(self._allreduce_ops[metric])

        # Override the reduced values back into logs dictionary
        # for other callbacks to use.
        for metric, value in reduced_logs.items():
            logs[metric] = value / self._allreduce_ranks

    def _horovod_average_metrics_in_place(self, logs):
        logs = logs or {}
        reduced_logs = {}
        import horovod.tensorflow as hvd

        if self._allreduce_ranks <= 1.:
            self._allreduce_ranks = float(hvd.size())
        # Reduce every metric among workers. Sort metrics by name
        # to ensure consistent order.
        for metric, value in sorted(logs.items()):
            from tensorflow.python.eager import context
            if context.executing_eagerly():
                reduced_logs[metric] = hvd.allreduce(K.constant(value, name=metric)).numpy()
            else:
                if metric not in self._m_vars:
                    with K.name_scope('MetricAverageCallback'):
                        var = K.variable(value, name=metric)
                        K.get_session().run(var.initializer)
                        self._m_vars[metric] = var
                        self._allreduce_ops[metric] = hvd.allreduce(var, device_dense=self._device)
                else:
                    K.set_value(self._m_vars[metric], value)
                reduced_logs[metric] = K.get_session().run(self._allreduce_ops[metric])
        # Override the reduced values back into logs dictionary
        # for other callbacks to use.
        for metric, value in reduced_logs.items():
            logs[metric] = value

    def custom_on_train_batch_end(self, step, logs=None):
        self._accumulated_time_secs += time.time() - self._last_triggered_time
        if step % self._summary_steps == 0:
            if self._strategy == "horovod":
                self._horovod_average_metrics_in_place(logs)
            elif self._strategy == "byteps":
                self._byteps_average_metrics_in_place(logs)
            if self._should_summary:
                secs_per_step = self._accumulated_time_secs / float(self._summary_steps)
                steps_per_sec = float(self._summary_steps) / self._accumulated_time_secs
                logging.info("Update %d\tTrainingLoss=%.2f\tSpeed %.3f secs/step %.1f steps/sec",
                             step, logs["loss"], secs_per_step, steps_per_sec)
                logging_metrics = {"step": step}
                if self._lr_schedule is not None:
                    logging_metrics["lr"] = self._lr_schedule(step - 1).numpy()
                for metric, value in logs.items():
                    reduction = get_metric_reduction(metric)
                    if reduction == METRIC_REDUCTION.MEAN:
                        logging_metrics[metric] = value
                    else:
                        assert reduction == METRIC_REDUCTION.SUM, (
                            f"Unknown reduction type of {metric}: {type(metric)}")
                        if self._last_triggered_metric is None:
                            self._last_triggered_metric = {}
                        this_metric_value = value - self._last_triggered_metric.get(metric, 0.)
                        this_metric_value *= self._allreduce_ranks
                        self._last_triggered_metric[metric] = value
                        logging_metrics[metric + "_per_step"] = this_metric_value / float(self._summary_steps)
                        logging_metrics[metric + "_per_sec"] = this_metric_value / self._accumulated_time_secs
                logging.info(logging_metrics)
                logging_metrics.pop("step")
                try:
                    for metric, value in logging_metrics.items():
                        tf.summary.scalar(compat.GlobalKeys.TBPREFIX_TRAINING + f"/{metric}", value, step=step)
                except tf.errors.OpError:
                    logging.info("Fail to summary metrics.")
            self._accumulated_time_secs = 0.



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import os
import re
import time
import traceback

import numpy
import tensorflow as tf
from absl import logging

from neurst.utils import compat
from neurst.utils.converters import Converter, build_converter


def remove_checkpoint_by_prefix(dirname, prefix):
    if prefix is None:
        return
    prefix = os.path.join(dirname, prefix)
    datas = tf.io.gfile.glob(prefix + ".data-?????-of-?????")
    for f in datas + [prefix + ".index"]:
        try:
            tf.io.gfile.remove(f)
        except tf.errors.OpError:
            logging.info(traceback.format_exc())


def restore_custom_checkpoint(checkpoint, checkpoint_path, model):
    """ Restore checkpoint from checkpoint_path.

    Args:
        checkpoint: A tf.train.Checkpoint.
        checkpoint_path: A string indicating the checkpoint path.
        model: A keras model.

    Returns: The checkpoint path if successfully restored or None otherwise.
    """
    traced_vars = model.weights
    pre_vars = dict(
        [(v.name.split(":")[0], v.numpy()) for v in traced_vars])
    try:
        checkpoint.restore(checkpoint_path).expect_partial()
    except (tf.errors.OpError, ValueError):
        logging.info(traceback.format_exc())
        return None
    logging.info('Restoring checkpoint from {latest_ckpt}'.format(
        latest_ckpt=checkpoint_path))
    after_vars = dict(
        [(v.name.split(":")[0], v.numpy()) for v in traced_vars])
    restored_var_names = []
    unrestored_var_names = []
    for v_name, v in pre_vars.items():
        try:
            if numpy.sqrt(numpy.sum((v - after_vars[v_name]) ** 2)) < 1e-6:
                unrestored_var_names.append(v_name)
            else:
                restored_var_names.append(v_name)
        except TypeError:
            logging.info(f"Ignore non-numeric variable: {v_name}")
    if len(unrestored_var_names) == 0:
        logging.info("All variables matched with checkpoint: {}".format(checkpoint_path))
    elif len(restored_var_names) == 0:
        logging.info("No variables matched with checkpoint: {}".format(checkpoint_path))
        if model is not None:
            logging.info("Trying `keras_model.load_weights()`")
            try:
                model.load_weights(checkpoint_path)
            except (ImportError, ValueError, tf.errors.OpError, AssertionError):
                logging.info("Fail to call model.load_weights.")
                logging.info(traceback.format_exc())
                return None
    else:
        for v_name in restored_var_names:
            logging.info("Restored {}".format(v_name))
        for v_name in unrestored_var_names:
            logging.info("Unrestored {}".format(v_name))
    return checkpoint_path


class _CustomSaver(object):
    """ Custom checkpoint manager for saving checkpoints. """

    def __init__(self, directory, checkpoint, max_to_keep=8):
        """ Initializes the checkpoint manager.

        Args:
            directory: The path to a directory in which to write checkpoints.
            checkpoint: A checkpoint
            max_to_keep: The maximum checkpoint numbers to keep.

        Raises:
            ValueError: Neither `traced_vars` nor `model` is provided.
        """
        self._directory = directory
        if not tf.io.gfile.exists(self._directory):
            try:
                tf.io.gfile.makedirs(self._directory)
            except tf.errors.OpError:
                pass
        self._checkpoint = checkpoint
        self._max_to_keep = max_to_keep
        # a list of tuple: (checkpoint name, timestamp)
        self._all_model_checkpoints = []

    @property
    def checkpoint(self):
        return self._checkpoint

    @property
    def directory(self):
        return self._directory

    def _update_checkpoint_meta(self):
        """Updates the checkpoint file under each model dir. """
        while len(self._all_model_checkpoints) > self._max_to_keep:
            prefix, _ = self._all_model_checkpoints.pop(0)
            remove_checkpoint_by_prefix(dirname=self.directory, prefix=prefix)
        meta_data_str = "model_checkpoint_path: \"{}\"\n".format(self._all_model_checkpoints[-1][0])
        for path, _ in self._all_model_checkpoints:
            meta_data_str += "all_model_checkpoint_paths: \"{}\"\n".format(path)
        for _, timestamp in self._all_model_checkpoints:
            meta_data_str += "all_model_checkpoint_timestamps: {}\n".format(str(timestamp))
        with tf.io.gfile.GFile(os.path.join(self.directory, "checkpoint.incomplete"), "w") as fw:
            fw.write(meta_data_str)
        tf.io.gfile.rename(os.path.join(self.directory, "checkpoint.incomplete"),
                           os.path.join(self.directory, "checkpoint"),
                           overwrite=True)

    def save(self, prefix):
        output = self._checkpoint.write(os.path.join(self.directory, prefix))
        return output


class NameBasedCheckpointManager(_CustomSaver):
    """ The name-based checkpoint manager for saving and restoring variables. """

    def __init__(self, model, directory, max_to_keep=8, checkpoint_name="ckpt"):
        """ Initializes a custom checkpoint manager.

        Args:
            model: A tf.keras.Model.
            directory: The path to a directory in which to write checkpoints.
            max_to_keep: The maximum checkpoint numbers to keep.
            checkpoint_name: The name of each checkpoint.
        """
        self._model = model
        super(NameBasedCheckpointManager, self).__init__(
            directory=directory, checkpoint=tf.train.Checkpoint(
                **dict([(x.name.split(":")[0], x) for x in model.weights])),
            max_to_keep=max_to_keep)
        self._checkpoint_name = checkpoint_name
        logging.info("Creates checkpoint manager for directory: {}".format(directory))

    def restore(self, restore_path=None):
        """ Restores checkpoint from `save_path` or self._directory by default. """
        if restore_path is None:
            restore_path = self.directory
        latest_ckpt = tf.train.latest_checkpoint(restore_path)
        if latest_ckpt is None:
            latest_ckpt = restore_path
        if latest_ckpt:
            return restore_custom_checkpoint(self.checkpoint, latest_ckpt, self._model)

    def save(self, checkpoint_number):
        prefix = "{}-{}".format(self._checkpoint_name, checkpoint_number)
        output = super(NameBasedCheckpointManager, self).save(prefix)
        self._all_model_checkpoints.append((prefix, time.time()))
        self._update_checkpoint_meta()
        return output


class KeepBestCheckpointSaver(_CustomSaver):
    """ Custom Checkpoint manager for saving and restoring variables. """

    def __init__(self, model, directory, metric, max_to_keep=8, checkpoint_name="ckpt"):
        """ Initializes a custom checkpoint manager.

        Args:
            model: A keras model.
            directory: The path to a directory in which to write checkpoints.
            metric: A metric object.
            max_to_keep: The maximum checkpoint numbers to keep.
            checkpoint_name: The name of each checkpoint.
        """
        if directory is None:
            directory = compat.get_saver_or_default().directory
            if not directory.endswith("/"):
                directory += "/"
            directory += "best"
        super(KeepBestCheckpointSaver, self).__init__(
            checkpoint=tf.train.Checkpoint(**dict([(x.name.split(":")[0], x) for x in model.weights])),
            directory=directory, max_to_keep=max_to_keep)
        self._metric = metric
        self._checkpoint_name = checkpoint_name
        logging.info("Creates custom keep-best checkpoint manager for directory: {}".format(directory))

    def save(self, checkpoint_number, metric_value):
        """ Saves a checkpoint and updates meta values if `metric_value` is better.

        Args:
            checkpoint_number: The current step.
            metric_value: The metric result.

        Returns:
            The path to the checkpoint if it is saved, otherwise None.

        """
        # whether to save or not
        if (0 <= len(self._all_model_checkpoints) < self._max_to_keep
            or self._metric.greater_or_eq(metric_value, self._all_model_checkpoints[0][1])):
            saved_prefix = "{}-{}-{}".format(self._checkpoint_name, checkpoint_number,
                                             ("%.2f" % self._metric.get_value(metric_value)))
            path = super(KeepBestCheckpointSaver, self).save(saved_prefix)
            self._all_model_checkpoints.append((saved_prefix, self._metric.get_value(metric_value)))
            self._all_model_checkpoints = sorted(
                self._all_model_checkpoints,
                key=functools.cmp_to_key(lambda x, y: (-int(self._metric.greater_or_eq(y[1], x[1])
                                                            and y[1] != x[1]))))
            self._update_checkpoint_meta()
            return path

        return None


class AverageCheckpointSaver(_CustomSaver):
    """ Custom Checkpoint manager for saving averaged variables. """

    def __init__(self, model, directory, metric, max_to_keep=8, checkpoint_name="ckpt"):
        """ Initializes a custom checkpoint manager.

        Args:
            model: A keras model.
            directory: The path to a directory in which to write checkpoints.
            metric: A metric object.
            max_to_keep: The maximum checkpoint numbers to keep.
            checkpoint_name: The name of each checkpoint.
        """
        if directory is None:
            directory = compat.get_saver_or_default().directory
            if not directory.endswith("/"):
                directory += "/"
            directory += "best_avg"
        self._checkpoint_name = checkpoint_name
        self._traced_vars = dict([(x.name.split(":")[0], x) for x in model.weights])
        self._traced_var_names = list(self._traced_vars.keys())
        self._traced_var_numpys = []
        self._metric = metric

        v_numpys = dict([(n, v.numpy()) for n, v in self._traced_vars.items()])
        with tf.distribute.OneDeviceStrategy(device="/cpu:0").scope():
            self._avg_traced_vars = dict([(n, tf.Variable(v, dtype=v.dtype, name=n + "_avg"))
                                          for n, v in v_numpys.items()])
        super(AverageCheckpointSaver, self).__init__(
            directory=directory, max_to_keep=max_to_keep,
            checkpoint=tf.train.Checkpoint(**self._avg_traced_vars))
        logging.info("Create checkpoint manager for averaged checkpoint "
                     "of the latest {} checkpoints to dir: {}".format(self._max_to_keep, self.directory))

    def _average_checkpoint(self):
        """ Averages the checkpoints. """
        for idx, name in enumerate(self._traced_var_names):
            self._avg_traced_vars[name] = self._avg_traced_vars[name].assign(
                numpy.average([var_numpys[idx] for var_numpys in self._traced_var_numpys], axis=0))

    def save(self, checkpoint_number, metric_value):
        """ Saves a checkpoint and updates meta values if `metric_value` is better.

        Args:
            checkpoint_number: The current step.
            metric_value: The metric result.

        Returns:
            The path to the checkpoint if it is saved, otherwise None.

        """
        # keep the latest checkpoints
        self._traced_var_numpys.append(
            [self._traced_vars[x].numpy() for x in self._traced_var_names])
        if len(self._traced_var_numpys) > self._max_to_keep:
            self._traced_var_numpys.pop(0)
        if (0 <= len(self._all_model_checkpoints) < self._max_to_keep
            or self._metric.greater_or_eq(metric_value, self._all_model_checkpoints[0][1])):
            # Averages the checkpoints.
            for idx, name in enumerate(self._traced_var_names):
                self._avg_traced_vars[name] = self._avg_traced_vars[name].assign(
                    numpy.average([var_numpys[idx] for var_numpys in self._traced_var_numpys], axis=0))
            saved_prefix = "{}-{}-{}".format(self._checkpoint_name, checkpoint_number,
                                             ("%.2f" % self._metric.get_value(metric_value)))
            path = super(AverageCheckpointSaver, self).save(saved_prefix)
            self._all_model_checkpoints.append((saved_prefix, self._metric.get_value(metric_value)))
            self._all_model_checkpoints = sorted(
                self._all_model_checkpoints,
                key=functools.cmp_to_key(lambda x, y: (-int(self._metric.greater_or_eq(y[1], x[1])
                                                            and y[1] != x[1]))))
            self._update_checkpoint_meta()
            return path

        return None


def checkpoint_scope_name(checkpoint_path):
    """ Lists checkpoint variables and extract top scope name.

    Args:
        checkpoint_path: A string, the checkpoint path.

    Returns: A string or None.
    """
    var_names = [compat.wrapper_var_name(x[0]) for x in tf.train.list_variables(checkpoint_path)]
    prefixs = set()
    for n in var_names:
        n_tokens = n.strip().split("/")
        if len(n_tokens) > 1:
            prefixs.add(n_tokens[0])
    if len(prefixs) > 1:
        logging.info("WARNING: more than one scope names({}) extracted from {}. "
                     "Be careful to this behavior, "
                     "which may lead to unknown issues.".format(prefixs, checkpoint_path))
    return prefixs.pop()


def restore_checkpoint_if_possible(model, model_dir, var_name_pattern=None):
    """ Restores checkpoint from `model_dir` if exists. """
    latest_ckpt_path = tf.train.latest_checkpoint(model_dir)
    if not latest_ckpt_path:
        latest_ckpt_path = model_dir
        try:
            tf.train.list_variables(latest_ckpt_path)
        except (tf.errors.NotFoundError, ValueError, tf.errors.DataLossError):
            return None

    ckpt_scope_name = checkpoint_scope_name(latest_ckpt_path)
    vars = model.weights
    if var_name_pattern is None:
        checkpoint = tf.train.Checkpoint(
            **dict([(ckpt_scope_name + x.name.split(":")[0][x.name.index("/"):], x) for x in vars]))
    else:
        logging.info("Variables only match the {} will be restored.".format(var_name_pattern))
        checkpoint = tf.train.Checkpoint(
            **dict([(ckpt_scope_name + x.name.split(":")[0][x.name.index("/"):], x) for x in vars
                    if re.search(var_name_pattern, x.name) is not None]))
    return restore_custom_checkpoint(checkpoint, latest_ckpt_path, model)


def restore_checkpoint_if_possible_v2(model, path, model_name=None, from_prefix=None,
                                      to_prefix=None, name_pattern=None):
    """ Restores checkpoint.

    Args:
        model: A keras model.
        path: The path to the neurst checkpoint or the path/key for the converter.
        model_name: The converter name for converting checkpoints.
        from_prefix: The name prefix.
        to_prefix: The target name prefix.
        name_pattern: A regex.

    Returns: The ckpt path if successfully restored else None.
    """
    if not (model_name or from_prefix or to_prefix):
        return restore_checkpoint_if_possible(model, path, name_pattern)
    logging.info(f"Loading {model_name} ({path}).")
    converter: Converter = build_converter(model_name)
    tmp_ckpt = "ram://tmp_ckpt"
    converter.convert(path, tmp_ckpt)
    latest_ckpt_path = tf.train.latest_checkpoint(tmp_ckpt)
    if from_prefix is None:
        from_prefix = checkpoint_scope_name(latest_ckpt_path)
    else:
        from_prefix = from_prefix.strip("/")
    vars = model.weights
    if to_prefix is None:
        to_prefix = vars[0].split("/")[0]
    else:
        to_prefix = to_prefix.strip("/")
    if name_pattern is None:
        checkpoint = tf.train.Checkpoint(
            **dict([(x.name.split(":")[0].replace(to_prefix, from_prefix, 1), x) for x in vars]))
    else:
        logging.info("Variables only match the {} will be restored.".format(name_pattern))
        checkpoint = tf.train.Checkpoint(
            **dict([(x.name.split(":")[0].replace(to_prefix, from_prefix, 1), x) for x in vars
                    if re.search(name_pattern, x.name) is not None]))
    return restore_custom_checkpoint(checkpoint, latest_ckpt_path, model)



# coding: utf-8
import io
import os

from setuptools import find_packages, setup

NAME = "neurst"
DESCRIPTION = "Neural Speech Translation Toolkit"
URL = ""
EMAIL = "zhaochengqi.d@bytedance.com"
AUTHOR = "ZhaoChengqi"

# TODO: one must manually install following packages if needed
# TODO: one must manually install following packages if needed
ALTERNATIVE_REQUIRES = [
    "jieba>=0.42.1",  # unnecessary for all
    "subword-nmt>=0.3.7",  # unnecessary for all
    "thai-segmenter>=0.4.1",  # unnecessary for all
    "soundfile>=0.10",  # for speech processing
    "python_speech_features>=0.6",  # for speech processing
    "transformers>=3.4.0",  # Not necessary for all
    "sentencepiece>=0.1.7",
    "mecab-python3>=1.0.3",  # for sacrebleu[ja]
    "ipadic>=1.0.0",  # for sacrebleu[ja]
    "torch>=1.7.0",  # for converting models from fairseq
    "fairseq>=0.10.1",  # for converting models from fairseq
    "tensorflow_addons>=0.11.2",  # for group normalization
]

REQUIRES = ["six>=1.11.0,<2.0.0",
            "pyyaml>=3.13",
            "sacrebleu>=1.4.0",
            "regex>=2019.1.24",
            "sacremoses>=0.0.38",
            # "tensorflow>=2.3.0", # ONE must manually install tensorflow
            "tqdm>=0.46",
            ]

DEV_REQUIRES = ["flake8>=3.5.0,<4.0.0",
                "mypy>=0.620; python_version>='3.6'",
                "tox>=3.0.0,<4.0.0",
                "isort>=4.0.0,<5.0.0",
                "pytest>=4.0.0,<5.0.0"] + REQUIRES + ALTERNATIVE_REQUIRES

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except IOError:
    long_description = DESCRIPTION

about = {}
with io.open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)

setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    keywords="neurst",
    packages=find_packages(exclude=["docs", "tests"]),
    install_requires=REQUIRES,
    tests_require=[
        "pytest>=4.0.0,<5.0.0"
    ],
    python_requires=">=3.6",
    extras_require={
        "dev": DEV_REQUIRES,
    },
    package_data={
        # for PEP484 & PEP561
        NAME: ["py.typed", "*.pyi"],
    },
    entry_points={
        "console_scripts": [
            "neurst-run = neurst.cli.run_exp:cli_main",
            "neurst-view = neurst.cli.view_registry:cli_main",
            "neurst-vocab = neurst.cli.generate_vocab:cli_main"
        ],
    },
)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import logging

from neurst.criterions import register_criterion
from neurst.criterions.criterion import Criterion
from neurst.metrics.metric import MetricWrapper
from neurst.models.model_utils import input_length_to_nonpadding
from neurst.utils.flags_core import Flag


@register_criterion
class LabelSmoothedCrossEntropy(Criterion):

    def __init__(self, args):
        """ Initializes the cross entropy with label smoothing.

        Args:
            args: A dict of full parameters.
        """
        super(LabelSmoothedCrossEntropy, self).__init__()
        self._label_smoothing = args["label_smoothing"]
        if self._label_smoothing > 0:
            logging.info("Using LabelSmoothedCrossEntropy with label_smoothing={}".format(self._label_smoothing))

    @staticmethod
    def class_or_method_args():
        """ Returns a list of args for flag definition. """
        return [Flag("label_smoothing", dtype=Flag.TYPE.FLOAT,
                     default=0., help="The label smoothing constant.")]

    def reduce_loss(self, model_inp, model_out):
        """ Reduces loss tensor for training according to the model inputs
            and outputs.

        Returns: A float tensor.
        """
        nll_sum, n_samples, n_tokens = self(model_inp, model_out)
        return tf.reduce_sum(nll_sum) / tf.reduce_sum(n_tokens)

    def reduce_metrics(self, eval_res_list) -> dict:
        """ Reduces the metrics according to a list of returned value from `eval`.

        Args:
            eval_res_list: A list of tuples of numpy.ndarray generated by `self.__call__`
                and model.__call__.

        Returns:
            A dict of reduced metrics for evaluation.
        """
        nll_sum, nll_samples, nll_tokens = 0., 0., 0.
        for _nll_sum, _nll_samples, _nll_tokens in eval_res_list:
            nll_sum += tf.reduce_sum(_nll_sum)
            nll_samples += tf.reduce_sum(_nll_samples)
            nll_tokens += tf.reduce_sum(_nll_tokens)
        nll = nll_sum / nll_samples
        ppl = 2. ** (nll_sum / nll_tokens)
        return {"NLL": nll, "PPL": ppl}

    def as_metric(self):
        """ Returns a wrapper class of Metric. """
        return MetricWrapper(flag="NLL", greater_is_better=False)

    def __call__(self, model_inp, model_out):
        """ Calculates.

        Args:
            model_inp: A dict containing the model inputs.
            model_out: The logits tensor or a dict containing the logits tensor.
                The logits tensor with shape [batch, max_len, vocab_size].

        Returns:
            The (nll_sum, num_of_samples(batch), num_of_tokens) with shape:
            nll_sum: [batch_size, ]
            num_of_samples: [1, ],
            num_of_tokens: [batch_size, ]
        """
        logits = model_out
        if isinstance(model_out, dict):
            logits = model_out["logits"]
        elif not isinstance(model_out, tf.Tensor):
            raise ValueError("Not supported type of model_out: {}".format(type(model_out)))

        logits = tf.cast(logits, tf.float32)
        labels = model_inp["trg"]

        with tf.name_scope("loss"):
            vocab_size = logits.get_shape()[-1]
            confidence = 1.0 - self._label_smoothing
            low_confidence = self._label_smoothing / tf.cast(vocab_size - 1, tf.float32)
            soft_target = tf.one_hot(
                tf.cast(labels, tf.int32),
                depth=vocab_size,
                on_value=confidence,
                off_value=low_confidence)
            # this may cause NaN when meets bad sample
            xentropy = tf.nn.softmax_cross_entropy_with_logits(
                logits=logits, labels=soft_target)
            # xentropy = - tf.reduce_sum(soft_target * tf.nn.log_softmax(logits), axis=-1)
            # Calculate the best (lowest) possible value of cross entropy, and
            # subtract from the cross entropy loss.
            if self._label_smoothing:
                normalizing_constant = -(
                    confidence * tf.math.log(confidence)
                    + tf.cast(vocab_size - 1, tf.float32) * low_confidence
                    * tf.math.log(low_confidence + 1e-20))
                xentropy -= normalizing_constant
            # else:
            # TODO(ZhaoChengqi) https://github.com/tensorflow/tensorflow/issues/32578
            # xentropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            #     logits=logits, labels=labels)
            if "trg_padding" in model_inp:
                weights = tf.cast(1 - model_inp["trg_padding"], tf.float32)
            else:
                weights = input_length_to_nonpadding(model_inp["trg_length"], tf.shape(labels)[1], tf.float32)
            nll_sum = tf.reduce_sum(xentropy * weights, axis=1)
            n_samples = tf.cast(tf.expand_dims(tf.shape(labels)[0], axis=0), dtype=tf.float32)
            n_tokens = tf.reduce_sum(weights, axis=1)
            return nll_sum, n_samples, n_tokens



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.utils import compat


def input_padding_to_bias(input_padding):
    """ Creates a bias tensor according to the padding tensor for attention.

    Args:
        input_padding: A float tensor with shape [batch_size, max_length],
                indicating the padding positions, where 1.0 for padding and
                0.0 for non-padding.

    Returns:
        Attention bias tensor with shape [batch_size, max_length]
    """
    with tf.name_scope("attention_bias"):
        bias = input_padding * compat.FLOAT_MIN
    return bias


def lower_triangle_attention_bias(length, dtype=None):
    """ Create a bias tensor for decoder self attention.

      Allows a query to attend to all positions up to and including its own.
    Args:
        length: A scalar.

    Returns: A float Tensor of shape [1, 1, length, length], with FLOAT_MIN in
      padding positions and 0 in non-padding positions.

    """
    if dtype is None:
        dtype = compat.CUSTOM_GLOBAL_FLOATX
    with tf.name_scope("decoder_self_attention_bias"):
        lower_triangle = tf.cast(
            tf.linalg.band_part(tf.ones([length, length]), -1, 0), dtype=dtype)
        bias = tf.reshape(compat.FLOAT_MIN * (1. - lower_triangle),
                          [1, 1, length, length])
    return bias


def stack_beam_size(x, beam_size):
    """ Tiles a given tensor by beam_size.

    Args:
        x: A tensor with shape [batch_size, ...].
        beam_size: An int scalar.

    Returns:
        The tiled tensor with shape [batch_size * beam_size, ...]

    Raises:
        AssertionError: if the shape of tensor does not match
          [batch_size, 1, 1, timesteps] when tensor.ndims == 4.
        NotImplementedError: if tensor.ndims > 4.
    """
    assert isinstance(x, tf.Tensor)
    original_shape = tf.shape(x)
    x = tf.expand_dims(x, axis=1)
    tile_dims = [1] * x.shape.ndims
    tile_dims[1] = beam_size
    tiled_x = tf.tile(x, tile_dims)
    tiled_shape = tf.concat([[-1], original_shape[1:]], axis=0)
    return tf.reshape(tiled_x, tiled_shape)


def static_shape_list(tensor):
    """Return a list of the tensor's shape, and ensure no None values in list."""
    # Get statically known shape (may contain None's for unknown dimensions)
    shape = tensor.get_shape().as_list()

    # Ensure that the shape values are not None
    dynamic_shape = tf.shape(tensor)
    for i in range(len(shape)):  # pylint: disable=consider-using-enumerate
        if shape[i] is None:
            shape[i] = dynamic_shape[i]
    return shape


def static_tensorshape(tensor):
    """ Returns the static TensorShape. """
    return tf.TensorShape(tensor.get_shape().as_list())


def dynamic_tensorshape_except_last_dim(tensor):
    """ Returns a tf.TensorShape with only last dim having the static shape. """
    shape_list = static_shape_list(tensor)
    # Only the last
    for i in range(len(shape_list) - 1):
        shape_list[i] = None

    if isinstance(shape_list[-1], tf.Tensor):
        shape_list[-1] = None
    return tf.TensorShape(shape_list)


def one_entry_bias(on_entry, num_entries,
                   on_value, off_value, dtype=None):
    """ Builds a bias vector to be added to log_probs for special use.

    Args:
        on_entry: A python integer.
        num_entries: A python integer.
        on_value: A scalar defining the value to fill in
            output when `index = on_entry`.
        off_value: A scalar defining the value to fill in
            output when `index != on_entry`.
        dtype: The tensor type.

    Returns: A bias vector with shape [num_entries, ].
    """
    if dtype is None:
        dtype = compat.CUSTOM_GLOBAL_FLOATX
    bias = tf.one_hot(
        [on_entry], num_entries,
        on_value=tf.convert_to_tensor(on_value, dtype=dtype),
        off_value=tf.convert_to_tensor(off_value, dtype=dtype),
        dtype=tf.dtypes.as_dtype(dtype))
    return tf.squeeze(bias, axis=0)


def tile_tensor(tensor, size, axis=0):
    """ Stacks a given tensor `size` times on a specific axis.

    For example, tensor=[1, 2, 3, 4], beam_size=3, axis=0 get the tensor
    [ [1, 2, 3, 4],
      [1, 2, 3, 4],
      [1, 2, 3, 4] ]

    tensor=[[1, 2, 3], [3, 4, 5]], beam_size=1, axis=1 get the tensor
    [ [[1, 2, 3]], [[3, 4, 5]] ]

    Args:
        tensor: A Tensor.
        size: A python integer, the size to be stacked.
        axis: A python integer.

    Returns: A Tensor.
    """
    tensor = tf.expand_dims(tensor, axis=axis)
    tile_dims = [1] * tensor.get_shape().ndims
    tile_dims[axis] = size
    return tf.tile(tensor, tile_dims)


def compute_batch_indices(batch_size, k):
    """ Computes the i'th coordinate that contains the batch index for gathers.

    Batch pos is a tensor like [[0,0,0,0,],[1,1,1,1],..]. It says which
    batch the beam item is in. This will create the i of the i,j coordinate
    needed for the gather.

    Args:
        batch_size: A python integer, the batch size.
        k: A python integer, the beam width.

    Returns: A Tensor.
    """
    # [beam_size, batch_size]: [[0, 1, 2,..., batch_size], [0, 1, 2,..., batch_size], ...]
    batch_pos = tile_tensor(tf.range(batch_size), k)
    batch_pos = tf.transpose(batch_pos)
    return batch_pos



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import os
import tarfile
import time
import zipfile
from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf
from absl import logging

from neurst.data.audio import FeatureExtractor, build_feature_extractor
from neurst.data.dataset_utils import glob_tfrecords, load_tfrecords
from neurst.data.datasets import Dataset, TFRecordDataset, register_dataset
from neurst.data.datasets.text_gen_dataset import TextGenDataset
from neurst.utils.compat import DataStatus
from neurst.utils.flags_core import Flag, ModuleFlag
from neurst.utils.misc import to_numpy_or_python_type

try:
    import soundfile
except (ImportError, OSError):
    pass


@six.add_metaclass(ABCMeta)
class RawAudioDataset(Dataset):
    """ Read from publicly available audio dataset.

    The raw dataset should yield element with following format:
        {
            "audio": ......,
            "transcript": ......,
            "translation": ......,
        }
    """

    def __init__(self, args):
        super(RawAudioDataset, self).__init__()
        self._input_tarball = args["input_tarball"]
        self._transcripts = None
        self._translations = None
        self._feature_extractor = build_feature_extractor(args)
        try:
            import soundfile
            _ = soundfile
        except ImportError:
            raise ImportError("Please install soundfile with: pip3 install soundfile")
        except OSError:
            raise OSError("sndfile library not found. Please install with: "
                          "apt-get install libsndfile1")

    @staticmethod
    def class_or_method_args():
        return [
            Flag("input_tarball", dtype=Flag.TYPE.STRING, default=None,
                 help="The original tarball."),
            ModuleFlag(FeatureExtractor.REGISTRY_NAME, default=None,
                       help="The audio feature extractor.")
        ]

    @contextlib.contextmanager
    def open_tarball(self, mode="tar"):
        if not tf.io.gfile.exists(self._input_tarball):
            raise ValueError(f"Input tarball not found: {self._input_tarball}")
        mode = mode.lower()
        if mode == "tar":
            fp = tf.io.gfile.GFile(self._input_tarball, "rb")
            tar = tarfile.open(fileobj=fp, mode="r:*")
            yield tar
            tar.close()
            fp.close()
        elif mode == "zip":
            local_path = self._input_tarball
            if self._input_tarball.startswith("hdfs://"):
                local_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                          os.path.basename(self._input_tarball))
                if not os.path.exists(local_path):
                    if os.path.exists(local_path + ".incomplete"):
                        while not os.path.exists(local_path):
                            time.sleep(20)
                    else:
                        logging.info(f"Copying ZIP file from HDFS to local disk {local_path}...")
                        tf.io.gfile.copy(self._input_tarball, local_path + ".incomplete", overwrite=True)
                        tf.io.gfile.rename(local_path + ".incomplete", local_path, overwrite=True)
            fp = zipfile.ZipFile(local_path, "r")
            yield fp
            fp.close()
        else:
            raise NotImplementedError(f"Unsupported type of input tarball: {self._input_tarball}.")

    def extract_audio_feature(self, sig=None, rate=None,
                              file=None, fileobj=None,
                              mode="wav"):
        """ Reads the audio data and extracts audio features.

        Args:
            signal: A 1-D or 2-D numpy array of either integer or float data-type, the audio data.
            rate: int, the sample rate.
            file: str, the file to read from.
            fileobj: file-like object, the file to read from.
            mode: The format of the audio file.

        Returns:

        """
        if sig is None:
            assert (file is None) ^ (fileobj is None), (
                "Only one of `file` and `fileobj` should be provided.")
            mode = mode.lower()
            if mode in ["flac", "wav"]:
                sig, rate = soundfile.read(file or fileobj, dtype='float32')
            else:
                raise NotImplementedError
        if self._feature_extractor is None:
            return sig
        return self._feature_extractor(sig, rate)

    @property
    def transcripts(self):
        if self._transcripts is None:
            if self._translations is None:
                self.load_transcripts()
        return self._transcripts

    @property
    def translations(self):
        if self._translations is None:
            if self._transcripts is None:
                self.load_transcripts()
        return self._translations

    @property
    def num_samples(self):
        if self.transcripts is not None:
            return len(self.transcripts)
        elif self.translations is not None:
            return len(self.translations)
        raise ValueError("Fail to get the number of samples.")

    @property
    @abstractmethod
    def status(self):
        raise NotImplementedError

    @abstractmethod
    def load_transcripts(self):
        """ Loads transcripts (and translations if exists). """
        raise NotImplementedError

    @abstractmethod
    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
            shard_id: Generator yields on the `shard_id`-th shard of the whole dataset.
            total_shards: The number of total shards.
        """
        raise NotImplementedError


@register_dataset("audio_tfrecord")
class AudioTFRecordDataset(TFRecordDataset, TextGenDataset):
    """ The TF Record Dataset for audio to text.
    The element spec must be
        {
            'audio': ...,
            'transcript': ...,
         }
    """

    def __init__(self, args):
        """ Initializes the dataset. """
        super(AudioTFRecordDataset, self).__init__(args)
        self._feature_key = args["feature_key"]
        self._transcript_key = args["transcript_key"]
        for x in tf.data.TFRecordDataset(glob_tfrecords(self._data_path)).take(1):
            example = tf.train.Example()
            example.ParseFromString(x.numpy())
            if len(example.features.feature[self._feature_key].float_list.value) > 0:
                self._audio_is_extracted = True
            elif len(example.features.feature[self._feature_key].int64_list.value) > 0:
                self._audio_is_extracted = False
            else:
                raise ValueError
            if len(example.features.feature[self._transcript_key].bytes_list.value) > 0:
                self._transcript_is_projected = False
            elif len(example.features.feature[self._transcript_key].int64_list.value) > 0:
                self._transcript_is_projected = True
            else:
                raise ValueError
        if not hasattr(self, "_audio_is_extracted"):
            raise ValueError(f"Fail to read {self._data_path}")

    @staticmethod
    def class_or_method_args():
        this_args = TFRecordDataset.class_or_method_args()
        this_args.extend([
            Flag("feature_key", dtype=Flag.TYPE.STRING, default="audio",
                 help="The key of the audio features in the TF Record."),
            Flag("transcript_key", dtype=Flag.TYPE.STRING, default="transcript",
                 help="The key of the audio transcript/translation in the TF Record."),
        ])
        return this_args

    @property
    def status(self):
        return {
            "audio": (DataStatus.PROJECTED if self._audio_is_extracted
                      else Dataset.RAW),  # audio
            "transcript": (DataStatus.PROJECTED if self._transcript_is_projected
                           else DataStatus.RAW),  # transcript
        }

    @property
    def fields(self):
        return {
            self._feature_key: (tf.io.VarLenFeature(tf.float32) if self._audio_is_extracted
                                else tf.io.VarLenFeature(tf.int64)),
            self._transcript_key: (tf.io.VarLenFeature(tf.int64) if self._transcript_is_projected
                                   else tf.io.VarLenFeature(tf.string)),
        }

    @property
    def targets(self):
        """ Returns a list of targets. """
        if self._targets is None:
            assert not self._transcript_is_projected
            gen = self.build_iterator(map_func=lambda x: x["transcript"])
            self._targets = [x for x in gen()]
        return self._targets

    def build(self, auto_shard=False, map_func=None, map_output_dtypes=None,
              shuffle=True) -> tf.data.Dataset:

        try:
            return load_tfrecords(
                self._data_path, shuffle=self._shuffle_dataset and shuffle,
                deterministic=(not shuffle),
                map_func=lambda x: x if map_func is None else map_func(x),
                auto_shard=auto_shard, name_to_features=self.fields,
                feature_name_mapping={self._feature_key: "audio", self._transcript_key: "transcript"})
        except AttributeError:

            logging.info("Call Dataset.from_generator for AudioTFRecord")

            def gen():
                for data in load_tfrecords(self._data_path, shuffle=self._shuffle_dataset and shuffle,
                                           deterministic=(not shuffle),
                                           auto_shard=auto_shard, name_to_features=self.fields,
                                           feature_name_mapping={self._feature_key: "audio",
                                                                 self._transcript_key: "transcript"}):
                    data = to_numpy_or_python_type(data, bytes_as_str=True)
                    if map_func is not None:
                        data = map_func(data)
                    yield data

            return tf.data.Dataset.from_generator(gen, output_types=map_output_dtypes)

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset. """

        def gen():
            ds = load_tfrecords(self._data_path, shuffle=False, auto_shard=False,
                                name_to_features=self.fields,
                                sharding_index=shard_id, num_shards=total_shards,
                                feature_name_mapping={self._feature_key: "audio",
                                                      self._transcript_key: "transcript"})
            for x in ds:
                data = to_numpy_or_python_type(x, bytes_as_str=True)
                if map_func is not None:
                    data = map_func(data)
                yield data

        return gen


@register_dataset("audio_triple_tfrecord")
class AudioTripleTFRecordDataset(TFRecordDataset, TextGenDataset):
    """ The TF Record Dataset for audio to text.
    The element spec must be
        {
            'audio': ...,
            'transcript': ...,
            'translation': ...
         }
    """

    def __init__(self, args):
        """ Initializes the dataset. """
        super(AudioTripleTFRecordDataset, self).__init__(args)
        self._feature_key = args["feature_key"]
        self._transcript_key = args["transcript_key"]
        self._translation_key = args["translation_key"]
        for x in tf.data.TFRecordDataset(glob_tfrecords(self._data_path)).take(1):
            example = tf.train.Example()
            example.ParseFromString(x.numpy())
            if len(example.features.feature[self._feature_key].float_list.value) > 0:
                self._audio_is_extracted = True
            elif len(example.features.feature[self._feature_key].int64_list.value) > 0:
                self._audio_is_extracted = False
            else:
                raise ValueError
            if len(example.features.feature[self._transcript_key].bytes_list.value) > 0:
                self._transcript_is_projected = False
            elif len(example.features.feature[self._transcript_key].int64_list.value) > 0:
                self._transcript_is_projected = True
            else:
                raise ValueError
            if len(example.features.feature[self._translation_key].bytes_list.value) > 0:
                self._translation_is_projected = False
            elif len(example.features.feature[self._translation_key].int64_list.value) > 0:
                self._translation_is_projected = True
            else:
                raise ValueError
        if not hasattr(self, "_audio_is_extracted"):
            raise ValueError(f"Fail to read {self._data_path}")
        self._transcripts = None

    @staticmethod
    def class_or_method_args():
        this_args = TFRecordDataset.class_or_method_args()
        this_args.extend([
            Flag("feature_key", dtype=Flag.TYPE.STRING, default="audio",
                 help="The key of the audio features in the TF Record."),
            Flag("transcript_key", dtype=Flag.TYPE.STRING, default="transcript",
                 help="The key of the audio transcript in the TF Record."),
            Flag("translation_key", dtype=Flag.TYPE.STRING, default="translation",
                 help="The key of the audio translation in the TF Record."),
        ])
        return this_args

    @property
    def status(self):
        return {
            "audio": (DataStatus.PROJECTED if self._audio_is_extracted
                      else Dataset.RAW),  # audio
            "transcript": (DataStatus.PROJECTED if self._transcript_is_projected
                           else DataStatus.RAW),  # transcript
            "translation": (DataStatus.PROJECTED if self._translation_is_projected
                            else DataStatus.RAW),  # transcript
        }

    @property
    def fields(self):
        return {
            self._feature_key: (tf.io.VarLenFeature(tf.float32) if self._audio_is_extracted
                                else tf.io.VarLenFeature(tf.int64)),
            self._transcript_key: (tf.io.VarLenFeature(tf.int64) if self._transcript_is_projected
                                   else tf.io.VarLenFeature(tf.string)),
            self._translation_key: (tf.io.VarLenFeature(tf.int64) if self._translation_is_projected
                                    else tf.io.VarLenFeature(tf.string)),
        }

    @property
    def targets(self):
        """ Returns a list of targets. """
        if self._targets is None:
            assert not self._translation_is_projected
            gen = self.build_iterator(map_func=lambda x: x["translation"])
            self._targets = [x for x in gen()]
        return self._targets

    @property
    def transcripts(self):
        """ Returns a list of transcripts. """
        if self._transcripts is None:
            assert not self._transcript_is_projected
            gen = self.build_iterator(map_func=lambda x: x["transcript"])
            self._transcripts = [x for x in gen()]
        return self._transcripts

    def build(self, auto_shard=False, map_func=None, map_output_dtypes=None,
              shuffle=True) -> tf.data.Dataset:

        try:
            return load_tfrecords(
                self._data_path, shuffle=self._shuffle_dataset and shuffle,
                deterministic=(not shuffle),
                map_func=lambda x: x if map_func is None else map_func(x),
                auto_shard=auto_shard, name_to_features=self.fields,
                feature_name_mapping={self._feature_key: "audio", self._transcript_key: "transcript",
                                      self._translation_key: "translation"})
        except AttributeError:

            logging.info("Call Dataset.from_generator for AudioTripleTFRecordDataset")

            def gen():
                for data in load_tfrecords(self._data_path, shuffle=self._shuffle_dataset and shuffle,
                                           deterministic=(not shuffle),
                                           auto_shard=auto_shard, name_to_features=self.fields,
                                           feature_name_mapping={self._feature_key: "audio",
                                                                 self._transcript_key: "transcript",
                                                                 self._translation_key: "translation"}):
                    data = to_numpy_or_python_type(data, bytes_as_str=True)
                    if map_func is not None:
                        data = map_func(data)
                    yield data

            return tf.data.Dataset.from_generator(gen, output_types=map_output_dtypes)

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset. """

        def gen():
            ds = load_tfrecords(self._data_path, shuffle=False, auto_shard=False,
                                name_to_features=self.fields,
                                sharding_index=shard_id, num_shards=total_shards,
                                feature_name_mapping={self._feature_key: "audio",
                                                      self._transcript_key: "transcript",
                                                      self._translation_key: "translation"})
            for x in ds:
                data = to_numpy_or_python_type(x, bytes_as_str=True)
                if map_func is not None:
                    data = map_func(data)
                yield data

        return gen



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import os

import tensorflow as tf
from absl import logging

from neurst.utils.compat import get_distributed_worker_setting
from neurst.utils.misc import deprecated, flatten_string_list

_MIN_BUCKET_BOUNDARY = 8
_BUCKET_BOUNDARY_SCALE = 1.1
_MAX_BUCKET_BOUNDARY = 256


def map_data_for_keras(dataset):
    """ Maps data for training.
        For TF v2, the 2nd parameter is omitted to make Keras training work.

    Args:
        dataset: A tf.data.Dataset object.

    Returns:
        A tf.data.Dataset object.
    """

    def _fn(*args):
        return (args,)

    return dataset.map(
        _fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)


@deprecated
def _batch_examples_by_token(dataset,
                             batch_size,
                             bucket_boundaries,
                             padding_values,
                             padding_length,
                             example_length_func,
                             drop_remainder=True,
                             num_replicas_in_sync=1):
    """Group examples by similar lengths, and return batched dataset.

    Each batch of similar-length examples are padded to the same length, and may
    have different number of elements in each batch, such that:
      group_batch_size * padded_length <= batch_size.

    This decreases the number of padding tokens per batch, which improves the
    training speed.

    Args:
        dataset: Dataset of unbatched examples.
        batch_size: Max number of tokens per batch of examples.
        bucket_boundaries: A list of integers of the boundaries of each bucket.
        padding_values: A tuple of constants for padding.
        padding_length: A list/tuple of padding length, which will be passed to padded_decode.
        example_length_func: A callable function, which deduces the input examples to the maximum length.
        drop_remainder: Whether the last batch should be dropped in the case it has fewer than batch_size.
        num_replicas_in_sync: The number of GPUs or other workers. We will generate
            global batches, and each global batch is equally divisible by number of replicas.

    Returns:
      Dataset of batched examples with similar lengths.
    """
    # Get min and max boundary lists for each example. These are used to calculate
    # the `bucket_id`, which is the index at which:
    # buckets_min[bucket_id] <= len(example) < buckets_max[bucket_id]
    # Note that using both min and max lists improves the performance.
    buckets_min = [0] + bucket_boundaries[:-1]
    buckets_max = bucket_boundaries

    # Create list of batch sizes for each bucket_id, so that
    # bucket_batch_size[bucket_id] * buckets_max[bucket_id] <= batch_size
    bucket_batch_sizes = [batch_size // x // num_replicas_in_sync * num_replicas_in_sync
                          for x in buckets_max]

    # bucket_id will be a tensor, so convert this list to a tensor as well.
    bucket_batch_sizes = tf.constant(bucket_batch_sizes, dtype=tf.int64)

    def example_to_bucket_id(examples):
        """Return int64 bucket id for this example, calculated based on length."""
        seq_length = tf.cast(example_length_func(examples), tf.int32)

        conditions_c = tf.logical_and(
            tf.less_equal(buckets_min, seq_length),
            tf.less(seq_length, buckets_max))
        bucket_id = tf.reduce_min(tf.where(conditions_c))
        return bucket_id

    def window_size_fn(bucket_id):
        """Return number of examples to be grouped when given a bucket id."""
        return bucket_batch_sizes[bucket_id]

    def batching_fn(bucket_id, grouped_dataset):
        """Batch and add padding to a dataset of elements with similar lengths."""
        bucket_batch_size = window_size_fn(bucket_id)

        # Batch the dataset and add padding so that all input sequences in the
        # examples have the same length, and all target sequences have the same
        # lengths as well. Resulting lengths of inputs and targets can differ.
        return grouped_dataset.padded_batch(
            bucket_batch_size, padding_length,
            padding_values=padding_values, drop_remainder=drop_remainder)

    return dataset.apply(tf.data.experimental.group_by_window(
        key_func=example_to_bucket_id,
        reduce_func=batching_fn,
        window_size=None,
        window_size_func=window_size_fn))


def create_batch_bucket_boundaries(max_length,
                                   min_boundary=_MIN_BUCKET_BOUNDARY,
                                   boundary_scale=_BUCKET_BOUNDARY_SCALE):
    """ Creates training batch bucket boundaries.

    Args:
        max_length: The maximum length of example in dataset.
        min_boundary: Minimum length in boundary.
        boundary_scale: Amount to scale consecutive boundaries in the list.

    Returns:
        A list of bucket boundaries.
    """
    # Create bucket boundaries list by scaling the previous boundary or adding 1
    # (to ensure increasing boundary sizes).
    bucket_boundaries = []
    x = min_boundary
    while x < max_length:
        bucket_boundaries.append(x)
        x = max(x + 1, int(x * boundary_scale))
    if bucket_boundaries[-1] < max_length + 1:
        bucket_boundaries = bucket_boundaries + [max_length + 1]
    return bucket_boundaries


def associated_bucket_boundaries(a, b):
    """ Creates training batch bucket boundaries.

    Args:
        a: A list of bucket boundaries.
        b: Another list of bucket boundaries.

    Returns:
        Two refactored lists of bucket boundaries with the same size.
    """
    length1 = len(a)
    length2 = len(b)
    if length1 == length2:
        return a, b
    elif length1 > length2:
        step_size1 = length1 * 1. / length2
        step_size2 = 1
    else:
        step_size1 = 1
        step_size2 = length2 * 1. / length1
    new_boundaries1 = []
    new_boundaries2 = []
    i = 1
    while i < min(length1, length2) + 1:
        new_boundaries1.append(a[int(math.ceil(i * step_size1)) - 1])
        new_boundaries2.append(b[int(math.ceil(i * step_size2)) - 1])
        i += 1

    return new_boundaries1, new_boundaries2


@deprecated
def load_from_tfrecord_and_auto_shard(features_file, shuffle=True,
                                      example_parse_fn=None, deterministic=True):
    """ Loads TFRecords and does autot-sharding according to worker num.

    Args:
        features_file: The TFRecords file path.
        shuffle: Whether to shuffle files.
        example_parse_fn: The example parse function for TF Record.
        deterministic: Whether the outputs need to be produced in deterministic order.

    Returns: A dataset.
    """
    _files = features_file.split(",")
    _features_files = []
    for _file in _files:
        if tf.io.gfile.isdir(_file):
            _features_files.append(os.path.join(_file, "*train*"))
        elif tf.io.gfile.exists(_file):
            _features_files.append(_file)
        else:
            _features_files.append(_file + "*")
    logging.info("Load TFRecords from {}".format(str(_features_files)))
    dataset = tf.data.Dataset.list_files(_features_files, shuffle=shuffle)
    # auto sharding
    worker_id, num_workers, strategy = get_distributed_worker_setting()
    if num_workers > 1 and strategy in ["horovod", "byteps"] and not shuffle:
        logging.info("Shard %d of the whole dataset(total %d workers).", worker_id, num_workers)
        dataset = dataset.shard(num_workers, worker_id)
    # Read files and interleave results.
    # When training, the order of the examples will be non-deterministic.
    options = tf.data.Options()
    options.experimental_deterministic = deterministic
    dataset = dataset.interleave(
        lambda f: tf.data.TFRecordDataset(f, buffer_size=128 * 1024 * 1024),
        cycle_length=10,
        num_parallel_calls=tf.data.experimental.AUTOTUNE).with_options(options)
    if example_parse_fn is None:
        return dataset
    return dataset.map(example_parse_fn,
                       num_parallel_calls=tf.data.experimental.AUTOTUNE)


def parse_tfexample(serialized_example, name_to_features,
                    name_mapping=None, map_func=None):
    """ Parses TF example from TF Record. """
    parsed = tf.io.parse_single_example(serialized_example, name_to_features)
    elements = {}
    for k, v in parsed.items():
        if name_mapping is None or k not in name_mapping:
            elements[k] = tf.sparse.to_dense(v)
        else:
            elements[name_mapping[k]] = tf.sparse.to_dense(v)
    if map_func is None:
        return elements
    return map_func(elements)


def glob_tfrecords(file_path):
    _files = flatten_string_list(file_path)
    _features_files = []
    for _file in _files:
        if tf.io.gfile.isdir(_file):
            _features_files.extend(tf.io.gfile.glob(os.path.join(_file, "*train*")))
        elif tf.io.gfile.exists(_file):
            _features_files.append(_file)
        else:
            _features_files.extend(tf.io.gfile.glob(_file + "*"))
    return _features_files


def load_tfrecords(file_path,
                   name_to_features,
                   shuffle=False,
                   deterministic=True,
                   feature_name_mapping=None,
                   map_func=None,
                   sharding_index=0,
                   num_shards=1,
                   auto_shard=False) -> tf.data.Dataset:
    """ Loads TFRecords and does autot-sharding according to worker num.

    Args:
        file_path: The TFRecords file path.
        name_to_features: A `dict` mapping feature keys to `FixedLenFeature` or
            `VarLenFeature` values.
        shuffle: Whether to shuffle files.
        deterministic: Whether the outputs need to be produced in deterministic order.
        feature_name_mapping: A dict that maps the names in `name_to_features` to aliases.
        map_func: A callable function to process the data.
        sharding_index: The manually defined index for sharding.
        num_shards: The manually defined number of shards operating in parallel.
        auto_shard: Automatically shard the TFRecord parts if True.

    Returns: A dataset.
    """
    _features_files = glob_tfrecords(file_path)
    shuffle = shuffle and (num_shards == 1)
    dataset = tf.data.Dataset.list_files(_features_files, shuffle=shuffle)
    if num_shards > 1:
        logging.info("Shard %d of the whole dataset(total %d workers).", sharding_index, num_shards)
        dataset = dataset.shard(num_shards, sharding_index)
    else:
        # auto sharding
        worker_id, num_workers, strategy = get_distributed_worker_setting()
        if num_workers > 1 and strategy in ["horovod", "byteps"] and not shuffle and auto_shard:
            logging.info("Shard %d of the whole dataset(total %d workers).", worker_id, num_workers)
            dataset = dataset.shard(num_workers, worker_id)
    logging.info("Loading TF Records from: ")
    for _f in dataset:
        logging.info(f"   {_f.numpy()}")
    # Read files and interleave results.
    # When training, the order of the examples will be non-deterministic.
    options = tf.data.Options()
    options.experimental_deterministic = deterministic
    dataset = dataset.interleave(
        lambda f: tf.data.TFRecordDataset(f, buffer_size=128 * 1024 * 1024),
        cycle_length=10,
        num_parallel_calls=tf.data.experimental.AUTOTUNE).with_options(options)

    if name_to_features is None:
        return dataset
    return dataset.map(lambda x: parse_tfexample(x, name_to_features, feature_name_mapping, map_func),
                       num_parallel_calls=tf.data.experimental.AUTOTUNE)


def clean_dataset_by_length(dataset, data_max_lengths):
    """ Filters empty datas, or datas exceeded max length. """
    logging.info(f"Filtering empty data and datas exceeded max length={data_max_lengths}")
    return dataset.filter(
        lambda data_sample: tf.reduce_all([
            (length == -1 or length is None
             or tf.less_equal(tf.size(data_sample[k]), length))  # filter by max length
            and (length == -1 or (length != -1 and tf.size(data_sample[k]) > 1))  # filter out empty lines
            for k, length in data_max_lengths.items()
        ]))


@deprecated
def batch_sequential_dataset(dataset,
                             padding_values,
                             example_length_func=None,
                             batch_size=None,
                             batch_size_per_gpu=None,
                             batch_by_tokens=False,
                             bucket_boundaries=None,
                             data_max_lengths=None,
                             shuffer_buffer=0,
                             drop_remainder=True,
                             num_replicas_in_sync=1):
    """ Calls padded_batch under special settings for sequential dataset.

    Args:
        dataset: A parallel dataset.
        padding_values: A list of padding values, will be passed to dataset.padded_batch.
        example_length_func: A callable function that takes a dict as input and returns
            the "length" of this data sample.
        batch_size: The number of sentences or word tokens according to `batch_by_tokens`.
        batch_size_per_gpu: The per-GPU batch size.
        batch_by_tokens: A bool, whether to batch the data by word tokens.
        bucket_boundaries: A list integers indicating the boundaries of the bucket when
            `batch_by_tokens` is True.
        data_max_lengths: The maximum length of training data, None or a list/tuple of
            integers with the the size as data samples. -1 indicates scalar data with
            no 'length' checking.
        shuffer_buffer: The buffer size for shuffling.
        drop_remainder: Whether the last batch should be dropped in the case it has fewer than batch_size.
        num_replicas_in_sync: The number of GPUs or other workers. We will generate global
                batches, and each global batch is equally divisible by number of replicas.

    Returns:
        The batched dataset.
    """
    if data_max_lengths is None:
        data_max_lengths = {k: None for k in padding_values}
    assert len(data_max_lengths) == len(padding_values)

    if example_length_func is None:
        def example_length_func(examples):
            return tf.reduce_max([
                tf.size(examples[k]) for k, length in data_max_lengths.items() if length != -1])

    if batch_size is None and batch_size_per_gpu is None:
        raise ValueError("Either `batch_size` or `batch_size_per_gpu` needs to be provided.")
    elif batch_size is not None and batch_size_per_gpu is not None:
        logging.info("Both `batch_size` and `batch_size_per_gpu` are provided, use `batch_size_per_gpu`.")
    if batch_size_per_gpu is not None:
        batch_size = int(batch_size_per_gpu * num_replicas_in_sync)
    logging.info("The global batch size is {}, with batch_by_tokens={}".format(batch_size, batch_by_tokens))
    # filter out empty lines
    dataset = clean_dataset_by_length(dataset, data_max_lengths)
    dynamic_padding_length = {k: ([] if length == -1 else [None])
                              for k, length in data_max_lengths.items()}

    if batch_by_tokens:
        # shuffle
        if shuffer_buffer:
            dataset = dataset.shuffle(buffer_size=shuffer_buffer)
        max_length = max(max([_len or 0 for _len in data_max_lengths.values()]), 0)

        if not max_length:
            logging.info("Using pre-defined max length={}".format(_MAX_BUCKET_BOUNDARY))
            max_length = _MAX_BUCKET_BOUNDARY
        logging.info("Final check of the max length of the training data. "
                     "Filter out whose length is larger than {}".format(max_length))
        dataset = dataset.filter(
            lambda data_sample: tf.reduce_all([
                (length == -1) or (length is None) or tf.size(data_sample[k]) <= max_length
                for k, length in data_max_lengths.items()]))
        if bucket_boundaries is None:
            bucket_boundaries = create_batch_bucket_boundaries(max_length)

        return _batch_examples_by_token(
            dataset,
            batch_size=batch_size,
            drop_remainder=drop_remainder,
            padding_values=padding_values,
            padding_length=dynamic_padding_length,
            bucket_boundaries=bucket_boundaries,
            example_length_func=example_length_func,
            num_replicas_in_sync=num_replicas_in_sync)
    else:
        # shuffle
        if shuffer_buffer:
            dataset = dataset.shuffle(buffer_size=shuffer_buffer)
        padding_length = dynamic_padding_length
    logging.info("The padding length of the dataset is {}".format(padding_length))
    dataset = dataset.padded_batch(
        int(batch_size // num_replicas_in_sync * num_replicas_in_sync),
        padding_length, drop_remainder=drop_remainder, padding_values=padding_values)
    return dataset


def adjust_batch_size(batch_size=None,
                      batch_size_per_gpu=None,
                      bucket_boundaries=None,
                      boundaries_reduce_to_length_fn=None,
                      num_replicas_in_sync=1,
                      verbose=True):
    if batch_size is None and batch_size_per_gpu is None:
        raise ValueError("At least one of the `batch_size` and `batch_size_per_gpu` should be provided.")
    elif batch_size is not None and batch_size_per_gpu is not None:
        logging.info("Both `batch_size` and `batch_size_per_gpu` are provided, use `batch_size_per_gpu`.")
    if batch_size_per_gpu is not None:
        batch_size = int(batch_size_per_gpu * num_replicas_in_sync)
    if bucket_boundaries is None:
        batch_size = int(batch_size // num_replicas_in_sync * num_replicas_in_sync)
        if verbose:
            logging.info(f"The global batch size is {batch_size} samples.")
        return batch_size
    logging.info(f"The global batch size is {batch_size} tokens.")
    bucket_batch_sizes = []
    try:
        i = 0
        while True:
            bucket_batch_sizes.append(
                int(batch_size // boundaries_reduce_to_length_fn({k: v[i] for k, v in bucket_boundaries.items()})
                    // num_replicas_in_sync * num_replicas_in_sync))
            i += 1

    except IndexError:
        pass
    return bucket_batch_sizes


def batch_examples_by_token(dataset,
                            bucket_boundaries,
                            bucket_batch_sizes,
                            padding_values,
                            example_length_func,
                            extra_padded_shapes=None,
                            drop_remainder=True):
    """Group examples by similar lengths, and return batched dataset.

    Each batch of similar-length examples are padded to the same length, and may
    have different number of elements in each batch, such that:
      group_batch_size * padded_length <= batch_size.

    This decreases the number of padding tokens per batch, which improves the
    training speed.

    Args:
        dataset: Dataset of unbatched examples.
        bucket_batch_sizes: Max number of tokens per batch of examples or a list of batch size for each bucket.
        bucket_boundaries: A list of integers of the boundaries of each bucket.
        padding_values: A tuple of constants for padding.
        example_length_func: A callable function, which deduces the input examples to the maximum length.
        extra_padded_shapes: A dict containing extra shapes (not included in bucket boundaries) for padding.
        drop_remainder: Whether the last batch should be dropped in the case it has fewer than batch_size.

    Returns:
      Dataset of batched examples with similar lengths.
    """
    cnt = 0
    try:
        logging.info("The details of batching logic:")
        while True:
            _batch = bucket_batch_sizes
            if isinstance(bucket_batch_sizes, list):
                _batch = bucket_batch_sizes[cnt]
            _bounds = {k: v[cnt] for k, v in bucket_boundaries.items()}
            logging.info(f"   - batch={_batch}, bucket boundary={_bounds}")
            cnt += 1
    except IndexError:
        logging.info(f"  Total {cnt} input shapes are compiled.")
    if not isinstance(bucket_batch_sizes, list):
        bucket_batch_sizes = [bucket_batch_sizes] * cnt
    # bucket_id will be a tensor, so convert this list to a tensor as well.
    bucket_batch_sizes = tf.constant(bucket_batch_sizes, dtype=tf.int64)
    bucket_boundaries = {k: tf.constant(v, dtype=tf.int32) for k, v in bucket_boundaries.items()}

    def example_to_bucket_id(examples):
        """Return int64 bucket id for this example, calculated based on length."""
        seq_length = example_length_func(examples)

        conditions_c = tf.reduce_all([
            tf.less_equal(v, bucket_boundaries[k])
            for k, v in seq_length.items()], axis=0)
        bucket_id = tf.reduce_min(tf.where(conditions_c))
        return bucket_id

    def window_size_fn(bucket_id):
        """Return number of examples to be grouped when given a bucket id."""
        return bucket_batch_sizes[bucket_id]

    def batching_fn(bucket_id, grouped_dataset):
        """Batch and add padding to a dataset of elements with similar lengths."""
        bucket_batch_size = window_size_fn(bucket_id)
        padded_shapes = {k: [v[bucket_id]] for k, v in bucket_boundaries.items()}
        if extra_padded_shapes:
            for k, v in extra_padded_shapes.items():
                padded_shapes[k] = v

        # Batch the dataset and add padding so that all input sequences in the
        # examples have the same length, and all target sequences have the same
        # lengths as well. Resulting lengths of inputs and targets can differ.
        return grouped_dataset.padded_batch(
            bucket_batch_size, padded_shapes,
            padding_values=padding_values,
            drop_remainder=drop_remainder)

    return dataset.apply(tf.data.experimental.group_by_window(
        key_func=example_to_bucket_id,
        reduce_func=batching_fn,
        window_size=None,
        window_size_func=window_size_fn))



import importlib
import os

from neurst.data.datasets.dataset import Dataset, TFRecordDataset
from neurst.utils.registry import setup_registry

build_dataset, register_dataset = setup_registry(Dataset.REGISTRY_NAME, base_class=Dataset,
                                                 verbose_creation=True)
_ = TFRecordDataset
models_dir = os.path.dirname(__file__)
for file in os.listdir(models_dir):
    path = os.path.join(models_dir, file)
    if not file.startswith('_') and not file.startswith('.') and file.endswith('.py'):
        model_name = file[:file.find('.py')] if file.endswith('.py') else file
        module = importlib.import_module('neurst.data.datasets.' + model_name)

importlib.import_module("neurst.data.datasets.audio")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

import tensorflow as tf
from absl import logging

from neurst.data.datasets import register_dataset
from neurst.data.datasets.audio.audio_dataset import RawAudioDataset
from neurst.utils.compat import DataStatus
from neurst.utils.flags_core import Flag


@register_dataset("Librispeech")
class LibriSpeech(RawAudioDataset):
    """
    LibriSpeech is a corpus of approximately 1000 hours of read English speech.
        Homepage: http://www.openslr.org/12
        The raw dataset contains 7 files:
            - train-clean-100.tar.gz
            - train-clean-360.tar.gz
            - train-other-500.tar.gz
            - dev-clean.tar.gz
            - dev-other.tar.gz
            - test-clean.tar.gz
            - test-other.tar.gz
    """

    def __init__(self, args):
        super(LibriSpeech, self).__init__(args)
        self._excluded_file = args["excluded_file"]
        self._excluded_list = None
        if self._excluded_file is not None:
            if not tf.io.gfile.exists(self._excluded_file):
                raise ValueError(f"`excluded_file` not found: {self._excluded_file}")
            with tf.io.gfile.GFile(self._excluded_file) as fp:
                self._excluded_list = [x.strip().lower() for x in fp]

        self._transcripts_dict = None

    @staticmethod
    def class_or_method_args():
        this_args = super(LibriSpeech, LibriSpeech).class_or_method_args()
        this_args.append(
            Flag("excluded_file", dtype=Flag.TYPE.STRING, default=None,
                 help="A file containing transcriptions "
                      "that would be removed in the LibriSpeech corpus."))
        return this_args

    @property
    def status(self):
        return {
            "audio": DataStatus.RAW,
            "transcript": DataStatus.RAW
        }

    def load_transcripts(self):
        """ Loads transcripts (and translations if exists). """
        if self._transcripts_dict is not None:
            return
        logging.info(f"Loading transcriptions from tarball: {self._input_tarball}")
        n = 0
        trans = {}
        level0 = set()
        level1_cnt = 0
        level2_cnt = 0
        excluded_count = 0
        excluded_str = ""
        if self._excluded_list is not None:
            excluded_str = " ".join(self._excluded_list)
        self._transcripts = []
        with self.open_tarball("tar") as tar:
            for tarinfo in tar:
                if not tarinfo.isreg():
                    continue
                n += 1
                if n % 10000 == 0:
                    logging.info("Scanned %d entries...", n)
                if not tarinfo.name.endswith(".trans.txt"):
                    continue
                level1_cnt += 1
                # The file LibriSpeech/dev-clean/3170/137482/3170-137482.trans.txt
                # will contain lines such as:
                # 3170-137482-0000 WITH AN EDUCATION WHICH OUGHT TO ...
                # 3170-137482-0001 I WAS COMPELLED BY POVERTY ...
                key = tarinfo.name.strip(".trans.txt")
                path0, path1 = key.split("/")[-1].split("-")
                level0.add(path0)
                f = tar.extractfile(tarinfo)
                this_dict = {}
                for line in f.readlines():
                    tid, txt = line.decode("utf-8").strip("\n").split(" ", 1)
                    txt_tokens = txt.split()
                    if txt in excluded_str:
                        excluded_count += 1
                        this_dict[tid] = ""
                    elif len(txt_tokens) > 10 and (
                        " ".join(txt_tokens[:len(txt_tokens) // 2]) in excluded_str
                        or " ".join(txt_tokens[len(txt_tokens) // 2:]) in excluded_str):
                        excluded_count += 1
                        this_dict[tid] = ""
                    else:
                        txt = txt.lower()
                        this_dict[tid] = txt
                        self._transcripts.append(txt)
                logging.info("[%s] = %d utterances.", key, len(this_dict))
                level2_cnt += len(this_dict)
                if path0 not in trans:
                    trans[path0] = dict()
                trans[path0][path1] = this_dict
                f.close()
        logging.info("Total %d directories, %d sub-directories, %d utterances, %d matched excluded file",
                     len(level0), level1_cnt, level2_cnt, excluded_count)
        # {'2277': {'149896': {'2277-149896-0000': "HE WAS IN A FEVERED STATE OF MIND OWING TO THE', ...}, ...}
        self._transcripts_dict = trans

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
        """
        if total_shards > 1:
            total_samples = self.num_samples
            samples_per_part = total_samples // total_shards
            range_begin = samples_per_part * shard_id
            if shard_id == total_shards - 1:
                range_end = total_samples + 1
                logging.info(f"Iterate on dataset from {range_begin} "
                             f"to the end (total {total_samples}).")
            else:
                range_end = range_begin + samples_per_part
                logging.info(f"Iterate on dataset from {range_begin} "
                             f"to {range_end} (total {total_samples}).")

        def gen():
            if self._transcripts_dict is None:
                self.load_transcripts()
            with self.open_tarball("tar") as tar:
                n = 0
                for tarinfo in tar:
                    if not tarinfo.isreg():
                        continue
                    if not tarinfo.name.endswith(".flac"):
                        continue

                    uttid = re.sub(".*/(.+)\\.flac", "\\1", tarinfo.name)
                    path0, path1, _ = uttid.strip().split("-")
                    this_trans = self._transcripts_dict[path0][path1][uttid]
                    if this_trans.strip() == "":
                        continue
                    n += 1
                    if total_shards > 1:
                        if n < range_begin:
                            continue
                        if n >= range_end:
                            break
                    f = tar.extractfile(tarinfo)
                    audio = self.extract_audio_feature(fileobj=f, mode="flac")
                    f.close()
                    data_sample = {
                        "audio": audio,
                        "transcript": this_trans
                    }
                    if map_func is None:
                        yield data_sample
                    else:
                        yield map_func(data_sample)

        return gen



import os
import tempfile

from neurst.data.text.vocab import Vocab

word_tokens = ["Hello", "World", "yes", "i", "I"]


def test():
    vocab = Vocab(word_tokens,
                  extra_tokens=["UNK", "EOS"])
    assert vocab._token_list == ["Hello", "World", "yes", "i", "I", "UNK", "EOS"]
    assert vocab.vocab_size == 7
    assert vocab.map_token_to_id(["Hello", "world", "man"],
                                 unknown_default=100) == [0, 100, 100]
    assert vocab.map_id_to_token([1, 0, 3]) == ["World", "Hello", "i"]

    vocab = Vocab(word_tokens,
                  extra_tokens=["UNK", "EOS"], lowercase=True)
    assert vocab._token_list == ["hello", "world", "yes", "i", "UNK", "EOS"]
    assert vocab.vocab_size == 6
    assert vocab.map_token_to_id(["Hello", "world", "man"],
                                 unknown_default=100) == [0, 1, 100]
    assert vocab.map_id_to_token([1, 0, 3]) == ["world", "hello", "i"]


def test_file():
    vocab_file = tempfile.NamedTemporaryFile(delete=False)
    with open(vocab_file.name, "w") as fw:
        for t in word_tokens:
            fw.write(t + "\t100\n")
    vocab = Vocab.load_from_file(vocab_file.name,
                                 extra_tokens=["UNK", "EOS"])
    assert vocab._token_list == ["Hello", "World", "yes", "i", "I", "UNK", "EOS"]
    assert vocab.vocab_size == 7
    assert vocab.map_token_to_id(["Hello", "world", "man"],
                                 unknown_default=100) == [0, 100, 100]
    assert vocab.map_id_to_token([1, 0, 3]) == ["World", "Hello", "i"]

    vocab = Vocab.load_from_file(vocab_file.name,
                                 extra_tokens=["UNK", "EOS"], lowercase=True)
    assert vocab._token_list == ["hello", "world", "yes", "i", "UNK", "EOS"]
    assert vocab.vocab_size == 6
    assert vocab.map_token_to_id(["Hello", "world", "man", "EOS"],
                                 unknown_default=100) == [0, 1, 100, 5]
    assert vocab.map_id_to_token([1, 0, 3]) == ["world", "hello", "i"]
    os.remove(vocab_file.name)


if __name__ == "__main__":
    test()
    test_file()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from neurst.data.datasets import build_dataset, register_dataset
from neurst.data.datasets.dataset import Dataset
from neurst.utils.flags_core import Flag


@register_dataset("multi_dataset")
class MultipleDataset(Dataset):

    def __init__(self, args):
        """ Initializes the multiple dataset.

        Args:
            args: containing `multiple_dataset`, which is like
                {
                    "data0": { "dataset.class": "", "dataset.params": ""},
                    "data1": { "dataset.class": "", "dataset.params": ""},
                    ......
                ]
        """
        super(MultipleDataset, self).__init__()
        self._datasets = {name: build_dataset(dsargs)
                          for name, dsargs in args["multiple_datasets"].items()}
        self._sample_weights = dict()
        if args["sample_weights"]:
            assert isinstance(args["sample_weights"], dict)
        else:
            args["sample_weights"] = {}
        sum = 0.
        for name in self._datasets:
            self._sample_weights[name] = args["sample_weights"].get(name, 1.)
            sum += self._sample_weights[name]
        for name in self._datasets:
            self._sample_weights[name] /= sum

    @staticmethod
    def class_or_method_args():
        return [
            Flag("multiple_datasets", dtype=Flag.TYPE.STRING,
                 help="A dict of dataset class and parameters, "
                      "where the key is the dataset name and "
                      "the value is a dict of arguments for one dataset."),
            Flag("sample_weights", dtype=Flag.TYPE.FLOAT,
                 help="A dict of weights for averaging metrics, where the key "
                      "is the dataset name. 1.0 for each by default.")
        ]

    @property
    def status(self):
        raise NotImplementedError

    @property
    def sample_weights(self):
        return self._sample_weights

    @property
    def datasets(self):
        return self._datasets

    def build(self, *args, **kwargs):
        raise NotImplementedError("Call each dataset's build function instead.")

    def build_iterator(self, *args, **kwargs):
        raise NotImplementedError("Call each dataset's build function instead.")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import traceback

from absl import logging

from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer

try:
    from transformers import AutoTokenizer
except ImportError:
    pass


@register_tokenizer("huggingface")
class HuggingFaceTokenizer(Tokenizer):

    def __init__(self, language, glossaries=None, **kwargs):
        super(HuggingFaceTokenizer, self).__init__(
            language=language, glossaries=glossaries, **kwargs)
        try:
            from transformers import AutoTokenizer
            _ = AutoTokenizer
        except ImportError:
            raise ImportError('Please install transformers with: pip3 install transformers')
        self._built = False
        self._codes = None

    def init_subtokenizer(self, codes):
        """ Lazily initializes huggingface tokenizer. """
        self._codes = codes

    def _lazy_init(self):
        codes = self._codes
        try:
            self._tokenizer = AutoTokenizer.from_pretrained(codes)
        except Exception as e:
            logging.info(traceback.format_exc())
            raise e
        self._built = True

    def tokenize(self, text, return_str=False):
        if not self._built:
            self._lazy_init()
            if not self._built:
                raise ValueError("call `init_subtokenizer` at first to initialize the tokenizer.")
        return self._output_wrapper(
            self._tokenizer.tokenize(self._convert_to_str(text)), return_str=return_str)

    def detokenize(self, text, return_str=True):
        if not self._built:
            self._lazy_init()
            if not self._built:
                raise ValueError("call `init_subtokenizer` at first to initialize the tokenizer.")
        return self._output_wrapper(
            self._tokenizer.convert_tokens_to_string(self._convert_to_list(text)), return_str=return_str)



import numpy
import pytest
import tensorflow as tf

from neurst.layers.attentions.multi_head_attention import MultiHeadAttention, MultiHeadSelfAttention


def test_multihead_attention():
    # num_input = 1
    # batch_size = 1
    # length = 2
    num_heads = 2
    num_units = 4
    dropout_rate = 0.
    output_depth = 3
    attention_layer = MultiHeadAttention(
        num_heads=num_heads,
        num_units=num_units,
        output_depth=output_depth,
        attention_dropout_rate=dropout_rate)
    query = tf.convert_to_tensor(
        [[[-1.3010577], [0.79014736]]], dtype=tf.float32)
    memory = tf.convert_to_tensor(
        [[[-1.1650294], [-0.88871276]]], dtype=tf.float32)
    # build layer
    attention_layer(query, memory=memory)
    predefined_weights = []
    for w in attention_layer.get_weights():
        # if "output_transform/kernel" in w.name:
        if w.shape == (4, 3):
            predefined_weights.append(
                numpy.array([[-0.18127477, -0.19210565, 0.6076416],
                             [0.43351638, 0.33151555, -0.49277717],
                             [-0.7911906, -0.48433813, 0.19314659],
                             [0.20843744, 0.76494586, -0.24448353]]))
        # elif "output_transform/bias" in w.name:
        elif w.shape == (3,):
            predefined_weights.append(
                numpy.array([0., 0., 0.]))
        # elif "q_transform/kernel" in w.name:
        elif w.shape == (1, 4):
            predefined_weights.append(
                numpy.array([[-0.8584202, 0.7467462, -0.9890406, 0.87380886]]))
        # elif "q_transform/bias" in w.name:
        elif w.shape == (4,):
            predefined_weights.append(
                numpy.array([0., 0., 0., 0.]))
        # elif "kv_transform/kernel" in w.name:
        elif w.shape == (1, 8):
            predefined_weights.append(
                numpy.array([[-0.741413, 0.05889565, -0.5399234, -0.5455819, 0.05704415,
                              0.75564325, -0.01531863, 0.7374855]]))
        # elif "kv_transform/bias" in w.name:
        elif w.shape == (8,):
            predefined_weights.append(
                numpy.array([0., 0., 0., 0., 0., 0., 0., 0.]))
    attention_layer.set_weights(predefined_weights)
    output = attention_layer(query, memory, is_training=False)
    assert numpy.sum((output.numpy() - numpy.array(
        [[[-0.5000115, -0.8363301, 0.5391713],
          [-0.49366233, -0.83081436, 0.5324018]]])) ** 2) < 1e-9


def test_multiheadself_attention():
    # num_input = 2
    # batch_size = 1
    # length = 2
    num_heads = 2
    num_units = 4
    dropout_rate = 0.
    output_depth = 3
    attention_layer = MultiHeadSelfAttention(
        num_heads=num_heads,
        num_units=num_units,
        output_depth=output_depth,
        attention_dropout_rate=dropout_rate)
    query = tf.convert_to_tensor(
        [[[-1.3010577, -2.3010577], [0.79014736, -0.79014736]]], dtype=tf.float32)
    bias = tf.convert_to_tensor([[-0.2276893, 0.11865579]], dtype=tf.float32)
    # build layer
    attention_layer(query)
    predefined_weights = []
    for w in attention_layer.get_weights():
        # if "output_transform/kernel" in w.name:
        if w.shape == (4, 3):
            predefined_weights.append(
                numpy.array([[-0.43040165, 0.841504, 0.8552141],
                             [-0.0111503, -0.14350098, -0.0815115],
                             [-0.54251707, 0.90105367, 0.7500602],
                             [-0.69162977, -0.10975248, 0.80686176]]))
        # elif "output_transform/bias" in w.name:
        elif w.shape == (3,):
            predefined_weights.append(
                numpy.array([0., 0., 0.]))
        # elif "qkv_transform/kernel" in w.name:
        elif w.shape == (2, 12):
            predefined_weights.append(
                numpy.array([[0.33632386, -0.5640344, -0.35587463, -0.5443644, -0.32000348,
                              -0.16645259, 0.38587856, 0.32863826, -0.5743795, 0.56257737,
                              0.17102379, -0.3622746],
                             [-0.1831755, -0.60358423, -0.55541337, 0.47493768, -0.37104,
                              -0.14766628, -0.64589596, 0.13060516, 0.4516778, -0.39557537,
                              0.08319885, 0.5473006]]))
        # elif "qkv_transform/bias" in w.name:
        elif w.shape == (12,):
            predefined_weights.append(
                numpy.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]))
    attention_layer.set_weights(predefined_weights)
    with pytest.raises(ValueError):
        attention_layer(query, bias=tf.convert_to_tensor(
            [[[-0.2276893, 0.11865579]]], dtype=tf.float32), is_training=False)
    output = attention_layer(query, bias=bias, is_training=False)
    assert numpy.sum((output.numpy() - numpy.array(
        [[[0.8281387, -0.55132246, -1.1976832],
          [0.8589912, -0.61944896, -1.2623004]]])) ** 2) < 1e-9


def test_multiheadself_attention_under_dec():
    # num_input = 2
    # length = 1
    num_heads = 2
    num_units = 4
    dropout_rate = 0.
    output_depth = 3
    batch_size = 1
    already_time = 2
    attention_layer = MultiHeadSelfAttention(
        num_heads=num_heads,
        num_units=num_units,
        output_depth=output_depth,
        attention_dropout_rate=dropout_rate)
    query = tf.convert_to_tensor(
        [[[-1.3010577, -2.3010577]]], dtype=tf.float32)
    # build layer
    attention_layer(query)
    predefined_weights = []
    for w in attention_layer.get_weights():
        # if "output_transform/kernel" in w.name:
        if w.shape == (4, 3):
            predefined_weights.append(
                numpy.array([[-0.43040165, 0.841504, 0.8552141],
                             [-0.0111503, -0.14350098, -0.0815115],
                             [-0.54251707, 0.90105367, 0.7500602],
                             [-0.69162977, -0.10975248, 0.80686176]]))
        # elif "output_transform/bias" in w.name:
        elif w.shape == (3,):
            predefined_weights.append(
                numpy.array([0., 0., 0.]))
        # elif "qkv_transform/kernel" in w.name:
        elif w.shape == (2, 12):
            predefined_weights.append(
                numpy.array([[0.33632386, -0.5640344, -0.35587463, -0.5443644, -0.32000348,
                              -0.16645259, 0.38587856, 0.32863826, -0.5743795, 0.56257737,
                              0.17102379, -0.3622746],
                             [-0.1831755, -0.60358423, -0.55541337, 0.47493768, -0.37104,
                              -0.14766628, -0.64589596, 0.13060516, 0.4516778, -0.39557537,
                              0.08319885, 0.5473006]]))
        # elif "qkv_transform/bias" in w.name:
        elif w.shape == (12,):
            predefined_weights.append(
                numpy.array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]))
    attention_layer.set_weights(predefined_weights)
    cache = {
        "keys":
            tf.reshape(tf.convert_to_tensor(numpy.array([[[-0.46546218, -1.0054358, 0.42906007, -1.6854379],
                                                          [1.078194, 1.1625745, -0.25033495, -1.980812]]]),
                                            dtype=tf.float32),
                       [batch_size, already_time, num_heads, num_units // num_heads]),
        "values":
            tf.reshape(tf.convert_to_tensor(numpy.array([[[-1.2360295, 0.69050753, -1.8204833, 0.23788007],
                                                          [2.3751693, -1.8772833, -0.2574517, 1.3010416]]]),
                                            dtype=tf.float32),
                       [batch_size, already_time, num_heads, num_units // num_heads])
    }
    attention_layer(query, cache=cache, is_training=False)
    output = attention_layer(query, cache=cache, is_training=False)
    assert numpy.sum((output.numpy() - numpy.array(
        [[[0.14431843, 0.4875261, 0.23705271]]])) ** 2) < 1e-9
    assert numpy.sum(
        (tf.reshape(cache["keys"], [batch_size, already_time + 2, num_units]).numpy()
         - numpy.array([[[-0.46546218, -1.0054358, 0.42906007, -1.6854379],
                         [1.078194, 1.1625745, -0.25033495, -1.980812],
                         [1.2701274, 0.5563531, 0.9841937, -0.72810733],
                         [1.2701274, 0.5563531, 0.9841937, -0.72810733]]]) ** 2)) < 1e-9
    assert numpy.sum(
        (tf.reshape(cache["keys"], [batch_size, already_time + 2, num_units]).numpy()
         - numpy.array([[[-1.2360295, 0.69050753, -1.8204833, 0.23788007],
                         [2.3751693, -1.8772833, -0.2574517, 1.3010416],
                         [-0.29203582, 0.17829615, -0.41395718, -0.78803015],
                         [-0.29203582, 0.17829615, -0.41395718, -0.78803015]]]) ** 2)) < 1e-9


if __name__ == "__main__":
    test_multihead_attention()
    test_multiheadself_attention()
    test_multiheadself_attention_under_dec()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six

from neurst.training.callbacks import CentralizedCallback
from neurst.utils.flags_core import Flag


@six.add_metaclass(ABCMeta)
class Validator(CentralizedCallback):
    REGISTRY_NAME = "validator"

    def __init__(self, args):
        super(Validator, self).__init__()
        self._eval_steps = args["eval_steps"]
        self._eval_start_at = args["eval_start_at"]

    @staticmethod
    def class_or_method_args():
        return [
            Flag("eval_steps", dtype=Flag.TYPE.INTEGER, default=1000,
                 help="The steps between two validation steps."),
            Flag("eval_start_at", dtype=Flag.TYPE.INTEGER, default=0,
                 help="The step to start validation process."),
        ]

    @abstractmethod
    def build(self, strategy, task, model):
        """ Builds the validator and returns self. """
        return self

    @abstractmethod
    def validate(self, step):
        """ Validation process. """
        raise NotImplementedError

    def custom_on_train_batch_end(self, step, logs=None):
        _ = logs
        if step >= self._eval_start_at and step % self._eval_steps == 0:
            self.validate(step)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import logging
import math
import re
import traceback

import numpy as np
import sacrebleu

from neurst.data.text.character import Character
from neurst.data.text.moses_tokenizer import MosesTokenizer
from neurst.data.text.thai_tokenizer import ThaiTokenizer
from neurst.metrics import register_metric
from neurst.metrics.metric import Metric


def bleu_count(hypothesis, references, max_n=4):
    ret_len_hyp = 0
    ret_len_ref = 0
    ret_clip_count = [0] * max_n
    ret_count = [0] * max_n
    for m in range(len(hypothesis)):
        hyp, ref = hypothesis[m], references[m]
        x = hyp.split()
        y = [r.split() for r in ref]
        x_len = len(x)
        y_len = [len(s) for s in y]
        n_ref = len(ref)

        closest_diff = 9999
        closest_length = 9999
        ref_ngram = dict()

        for i in range(n_ref):
            diff = abs(y_len[i] - x_len)
            if diff < closest_diff:
                closest_diff = diff
                closest_length = y_len[i]
            elif diff == closest_diff and y_len[i] < closest_length:
                closest_length = y_len[i]

            for n in range(max_n):
                sent_ngram = dict()
                for st in range(0, y_len[i] - n):
                    ngram = "%d" % (n + 1)
                    for k in range(n + 1):
                        j = st + k
                        ngram += " %s" % (y[i][j])
                    if ngram not in sent_ngram:
                        sent_ngram[ngram] = 0
                    sent_ngram[ngram] += 1
                for ngram in sent_ngram.keys():
                    if ngram not in ref_ngram or ref_ngram[ngram] < sent_ngram[ngram]:
                        ref_ngram[ngram] = sent_ngram[ngram]

        ret_len_hyp += x_len
        ret_len_ref += closest_length

        for n in range(max_n):
            hyp_ngram = dict()
            for st in range(0, x_len - n):
                ngram = "%d" % (n + 1)
                for k in range(n + 1):
                    j = st + k
                    ngram += " %s" % (x[j])
                if ngram not in hyp_ngram:
                    hyp_ngram[ngram] = 0
                hyp_ngram[ngram] += 1
            for ngram in hyp_ngram.keys():
                if ngram in ref_ngram:
                    ret_clip_count[n] += min(ref_ngram[ngram], hyp_ngram[ngram])
                ret_count[n] += hyp_ngram[ngram]

    return ret_clip_count, ret_count, ret_len_hyp, ret_len_ref


def corpus_bleu(hypothesis, references, max_n=4):
    clip_count, count, total_len_hyp, total_len_ref = bleu_count(hypothesis, references, max_n=max_n)
    brevity_penalty = 1.0
    bleu_scores = []
    for n in range(max_n):
        if count[n] > 0:
            bleu_scores.append(clip_count[n] / count[n])
        else:
            bleu_scores.append(0)
    if total_len_hyp < total_len_ref:
        brevity_penalty = math.exp(1 - total_len_ref / total_len_hyp)

    def my_log(x):
        if x == 0:
            return -9999999999.0
        elif x < 0:
            raise ValueError("x in log(x) must >= 0.")
        return math.log(x)

    log_bleu = 0.0
    for n in range(max_n):
        log_bleu += my_log(bleu_scores[n])
    bleu = brevity_penalty * math.exp(log_bleu / float(max_n))
    return [bleu] + bleu_scores, [brevity_penalty, total_len_hyp / total_len_ref, total_len_hyp, total_len_ref]


def sentence_bleu(hypothesis, references, max_n=4):
    clip_count, count, total_len_hyp, total_len_ref = bleu_count([hypothesis], [references], max_n=max_n)
    brevity_penalty = 1.0
    bleu_scores = []
    for n in range(max_n):
        bleu_scores.append((clip_count[n] + 0.01) / (count[n] + 0.01))  # smoothing
    if total_len_hyp < total_len_ref:
        brevity_penalty = math.exp(1 - total_len_ref / total_len_hyp)

    def my_log(x):
        if x == 0:
            return -9999999999.0
        elif x < 0:
            raise ValueError("x in log(x) must >= 0.")
        return math.log(x)

    log_bleu = 0.0
    for n in range(max_n):
        log_bleu += my_log(bleu_scores[n])
    bleu = brevity_penalty * math.exp(log_bleu / float(max_n))
    return [bleu] + bleu_scores, [brevity_penalty, total_len_hyp / total_len_ref, total_len_hyp, total_len_ref]


def incremental_bleu_count(hypothesis, references, max_n=4):
    ret_len_hyp = []
    ret_len_ref = []
    ret_clip_count = []
    ret_count = []
    for m in range(len(hypothesis)):
        hyp, ref = hypothesis[m], references[m]
        x = hyp.split()
        y = [r.split() for r in ref]
        x_len = len(x)
        y_len = [len(s) for s in y]
        n_ref = len(ref)

        ref_ngram = dict()

        for i in range(n_ref):
            for n in range(max_n):
                sent_ngram = dict()
                for st in range(0, y_len[i] - n):
                    ngram = "%d" % (n + 1)
                    for k in range(n + 1):
                        j = st + k
                        ngram += " %s" % (y[i][j])
                    if ngram not in sent_ngram:
                        sent_ngram[ngram] = 0
                    sent_ngram[ngram] += 1
                for ngram in sent_ngram.keys():
                    if ngram not in ref_ngram or ref_ngram[ngram] < sent_ngram[ngram]:
                        ref_ngram[ngram] = sent_ngram[ngram]
        y_len = sorted(y_len)
        ret_len_hyp.append([])
        ret_len_ref.append([])
        ret_clip_count.append([])
        ret_count.append([])

        hyp_ngram = dict()
        p_closest = 0
        for i in range(x_len):
            if i == 0:
                ret_clip_count[-1].append([0] * max_n)
                ret_count[-1].append([0] * max_n)
            else:
                ret_clip_count[-1].append(copy.deepcopy(ret_clip_count[-1][-1]))
                ret_count[-1].append(copy.deepcopy(ret_count[-1][-1]))

            j = i + 1
            ret_len_hyp[-1].append(i + 1)
            if j > y_len[p_closest]:
                while j > y_len[p_closest] and p_closest < n_ref - 1:
                    p_closest += 1
            tmp_closest_diff = 9999
            tmp_closest_len = 9999
            if p_closest > 0 and (j - y_len[p_closest - 1]) < tmp_closest_diff:
                tmp_closest_diff = j - y_len[p_closest - 1]
                tmp_closest_len = y_len[p_closest - 1]
            if p_closest < n_ref and (y_len[p_closest] - j) < tmp_closest_diff:
                tmp_closest_diff = y_len[p_closest] - j
                tmp_closest_len = y_len[p_closest]

            ret_len_ref[-1].append(tmp_closest_len)
            for n in range(max_n):
                st = i - n
                if st >= 0:
                    ngram = "%d" % (n + 1)
                    for k in range(n + 1):
                        j = st + k
                        ngram += " %s" % (x[j])
                    if ngram not in hyp_ngram:
                        hyp_ngram[ngram] = 0
                    hyp_ngram[ngram] += 1
                    ret_count[-1][-1][n] += 1
                    if ngram in ref_ngram and hyp_ngram[ngram] <= ref_ngram[ngram]:
                        ret_clip_count[-1][-1][n] += 1

    return ret_clip_count, ret_count, ret_len_hyp, ret_len_ref


def incremental_sent_bleu(hypothesis, references, max_n=4):
    clip_count, count, total_len_hyp, total_len_ref = incremental_bleu_count([hypothesis], [references], max_n=max_n)
    clip_count = clip_count[0]
    count = count[0]
    total_len_hyp = total_len_hyp[0]
    total_len_ref = total_len_ref[0]
    n_len = len(clip_count)
    ret = []
    for i in range(n_len):
        brevity_penalty = 1.0
        bleu_scores = []
        bleu = 0
        for n in range(max_n):
            if count[i][n] > 0:
                bleu_scores.append(clip_count[i][n] / count[i][n])
            else:
                bleu_scores.append(0)
        if total_len_hyp[i] < total_len_ref[i]:
            brevity_penalty = math.exp(1 - total_len_ref[i] / total_len_hyp[i])

        def my_log(x):
            if x == 0:
                return -9999999999.0
            elif x < 0:
                raise ValueError("x in log(x) must >= 0.")
            return math.log(x)

        log_bleu = 0.0
        for n in range(max_n):
            log_bleu += my_log(bleu_scores[n])
        bleu = brevity_penalty * math.exp(log_bleu / float(max_n))
        ret.append(bleu)
    return ret


def incremental_test_corpus_bleu(hypothesis, references, max_n=4):
    assert (len(hypothesis) == len(references))
    tmp_clip_count, tmp_count, tmp_total_len_hyp, tmp_total_len_ref = incremental_bleu_count(hypothesis, references,
                                                                                             max_n=max_n)
    clip_count = [0] * 4
    count = [0] * 4
    total_len_hyp = 0
    total_len_ref = 0
    for i in range(len(hypothesis)):
        for n in range(4):
            clip_count[n] += tmp_clip_count[i][-1][n]
            count[n] += tmp_count[i][-1][n]
        total_len_hyp += tmp_total_len_hyp[i][-1]
        total_len_ref += tmp_total_len_ref[i][-1]
    brevity_penalty = 1.0
    bleu_scores = []
    bleu = 0
    for n in range(max_n):
        if count[n] > 0:
            bleu_scores.append(clip_count[n] / count[n])
        else:
            bleu_scores.append(0)
    if total_len_hyp < total_len_ref:
        brevity_penalty = math.exp(1 - total_len_ref / total_len_hyp)

    def my_log(x):
        if x == 0:
            return -9999999999.0
        elif x < 0:
            raise ValueError("x in log(x) must >= 0.")
        return math.log(x)

    log_bleu = 0.0
    for n in range(max_n):
        log_bleu += my_log(bleu_scores[n])
    bleu = brevity_penalty * math.exp(log_bleu / float(max_n))
    return [bleu] + bleu_scores, [brevity_penalty, total_len_hyp / total_len_ref, total_len_hyp, total_len_ref]


def commonly_tokenize(s):
    """ Tokenizes sentence according to multi-bleu-detok.perl & mteval-v13a.pl """
    # language-independent part
    s = re.sub(r"-\n", r"", s)  # strip end-of-line hyphenation and join lines
    s = re.sub(r"\n", r" ", s)  # join lines
    s = re.sub(r"&quot;", r'"', s)  # convert SGML tag for quote to "
    s = re.sub(r"&amp;", r"&", s)  # convert SGML tag for ampersand to &
    s = re.sub(r"&lt;", r"<", s)  # convert SGML tag for less-than to >
    s = re.sub(r"&gt;", r">", s)  # convert SGML tag for greater-than to <
    s = " " + s + " "
    # language-dependent part (assuming Western languages):
    s = re.sub(r"([\{-~\[-` -&\(-\+:-@\/])", r" \1 ", s)  # tokenize punctuation
    s = re.sub(r"([^0-9])([\.,])", r"\1 \2 ", s)  # tokenize period and comma unless preceded by a digit
    s = re.sub(r"([\.,])([^0-9])", r" \1 \2", s)  # tokenize period and comma unless followed by a digit
    # s = re.sub(r"([\.,])([0-9])", r" \1 \2", s)  # tokenize period and comma unless followed by a digit
    s = re.sub(r"([0-9])(-)", r"\1 \2 ", s)  # tokenize dash when preceded by a digit

    return " ".join(s.strip().split())


ESCAPE_AMPERSAND = r'&', r'&amp;'
ESCAPE_PIPE = r'|', r'&#124;'
ESCAPE_LEFT_ANGLE_BRACKET = r'<', r'&lt;'
ESCAPE_RIGHT_ANGLE_BRACKET = r'>', r'&gt;'
ESCAPE_SINGLE_QUOTE = r"'", r"&apos;"
ESCAPE_DOUBLE_QUOTE = r'"', r'&quot;'
ESCAPE_LEFT_SQUARE_BRACKET = r"[", r"&#91;"
ESCAPE_RIGHT_SQUARE_BRACKET = r"]", r"&#93;"

ESCAPE_LIST = [ESCAPE_AMPERSAND,
               ESCAPE_PIPE,
               ESCAPE_LEFT_ANGLE_BRACKET,
               ESCAPE_RIGHT_ANGLE_BRACKET,
               ESCAPE_SINGLE_QUOTE,
               ESCAPE_DOUBLE_QUOTE,
               ESCAPE_LEFT_SQUARE_BRACKET,
               ESCAPE_RIGHT_SQUARE_BRACKET]


def unescape(s):
    for repl, patt in ESCAPE_LIST:
        s = re.sub(patt, repl, s)
    return s


@register_metric(["sacre_bleu",
                  "tok_bleu",
                  "detok_bleu",
                  "uncased_sacre_bleu",
                  "uncased_tok_bleu",
                  "uncased_detok_bleu"])
class BLEU(Metric):

    def __init__(self, language="en", *args, **kwargs):
        """ Initializes.

        Args:
            language: The language.
        """
        _ = args
        _ = kwargs
        super(BLEU, self).__init__()
        self._language = language
        if language in ["zh", "ja", "ko", "km"]:
            self._tokenize_fn = lambda x: Character.to_character(x, language=language)
        elif language == "th":
            tokenizer = ThaiTokenizer()
            self._tokenize_fn = lambda x: tokenizer.tokenize(x, return_str=True)
        else:
            tokenizer = MosesTokenizer(language=language)
            self._tokenize_fn = lambda x: tokenizer.tokenize(x, return_str=True)
        self._sacre_tokenize_str = "13a"
        if language == "zh":
            self._sacre_tokenize_str = "zh"
        elif language == "ja":
            self._sacre_tokenize_str = "ja-mecab"

    @staticmethod
    def _tokenize(ss, tok_fn, lc=False):
        assert isinstance(ss, list)
        if isinstance(ss[0], str):
            return [tok_fn(unescape(x.lower() if lc else x)) for x in ss]
        return [[tok_fn(unescape(x.lower() if lc else x)) for x in xx] for xx in ss]

    def set_groundtruth(self, groundtruth):
        """ Setup inside groundtruth.

        Args:
            groundtruth: A list of references,
                [sent0_ref, sent1_ref, ...]
                    or a list of multiple references,
                    [[sent0_ref0, sent1_ref0, ...],
                    [sent0_ref1, sent1_ref1, ...],
                    ......]
        """
        assert isinstance(groundtruth, list)
        if isinstance(groundtruth[0], str):
            groundtruth = [groundtruth]
        self._refs_for_sacre = groundtruth
        refs = list(map(list, zip(*groundtruth)))
        self._refs_for_tok = self._tokenize(refs, self._tokenize_fn, lc=False)
        self._uncased_refs_for_tok = self._tokenize(refs, self._tokenize_fn, lc=True)
        self._refs_for_tok = [
            [self._tokenize_fn(r) for r in rr] for rr in refs]
        self._uncased_refs_for_tok = [
            [self._tokenize_fn(r.lower()) for r in rr] for rr in refs]

    def tok_bleu(self, hypo, groundtruth=None, lc=False):
        tok_hypos = self._tokenize(hypo, self._tokenize_fn, lc=lc)
        if groundtruth is None:
            ref = self._uncased_refs_for_tok if lc else self._refs_for_tok
        else:
            if isinstance(groundtruth[0], str):
                groundtruth = [groundtruth]
            ref = self._tokenize(list(map(list, zip(*groundtruth))), self._tokenize_fn, lc=lc)
        try:
            bleu, _ = corpus_bleu(tok_hypos, ref)
            bleu = bleu[0]
        except IndexError:
            logging.info("Found empty lines.")
            print(traceback.format_exc())
            bleu = 0.
        except ZeroDivisionError:
            logging.info("Empty reference")
            print(traceback.format_exc())
            bleu = 0.
        return bleu * 100

    def detok_bleu(self, hypo, groundtruth=None, lc=False):
        # tok_hypos = self._tokenize(hypo, self._default_tokenize_fn, lc=lc)
        # if groundtruth is None:
        #     ref = self._uncased_refs_for_detok if lc else self._refs_for_detok
        # else:
        #     if isinstance(groundtruth[0], str):
        #         groundtruth = [groundtruth]
        #     ref = self._tokenize(list(map(list, zip(*groundtruth))), self._default_tokenize_fn, lc=lc)
        # try:
        #     bleu, _ = corpus_bleu(tok_hypos, ref)
        #     bleu = bleu[0]
        # except IndexError:
        #     logging.info("Found empty lines.")
        #     print(traceback.format_exc())
        #     bleu = 0.
        # except ZeroDivisionError:
        #     logging.info("Empty reference")
        #     print(traceback.format_exc())
        #     bleu = 0.
        # return bleu * 100
        if self._language in ["ko", "km", "th"]:
            return self.tok_bleu(hypo, groundtruth, lc)
        return self.sacre_bleu(hypo, groundtruth, lc)

    def sacre_bleu(self, hypo, groundtruth=None, lc=False):
        if groundtruth is None:
            ref = self._refs_for_sacre
        else:
            if isinstance(groundtruth[0], str):
                ref = [groundtruth]
            else:
                ref = groundtruth
        try:
            bleu = sacrebleu.corpus_bleu(
                hypo, ref, lowercase=lc, tokenize=self._sacre_tokenize_str)
            return bleu.score
        except IndexError:
            logging.info("Found empty lines.")
            print(traceback.format_exc())
            return 0.
        except ZeroDivisionError:
            logging.info("Empty reference")
            print(traceback.format_exc())
            return 0.

    def get_value(self, result):
        if isinstance(result, (float, np.float32, np.float64)):
            return result
        if self._flag in result:
            return result[self._flag]
        if self._flag.lower() in result:
            return result[self._flag.lower()]
        if self._language in ["ko", "km"]:
            return result["tok_bleu"]
        return result["sacre_bleu"]

    def call(self, hypothesis, groundtruth=None):
        """ Returns the BLEU result dict. """
        return {
            "sacre_bleu": self.sacre_bleu(hypothesis, groundtruth),
            "tok_bleu": self.tok_bleu(hypothesis, groundtruth),
            "detok_bleu": self.detok_bleu(hypothesis, groundtruth),
            "uncased_sacre_bleu": self.sacre_bleu(hypothesis, groundtruth, lc=True),
            "uncased_tok_bleu": self.tok_bleu(hypothesis, groundtruth, lc=True),
            "uncased_detok_bleu": self.detok_bleu(hypothesis, groundtruth, lc=True)}
